!=== COPYRIGHT AND LICENSE STATEMENT ===
!
!  This file is part of the TensorProductMultigrid code.
!  
!  (c) The copyright relating to this work is owned jointly by the
!  Crown, Met Office and NERC [2014]. However, it has been created
!  with the help of the GungHo Consortium, whose members are identified
!  at https://puma.nerc.ac.uk/trac/GungHo/wiki .
!  
!  Main Developer: Eike Mueller
!  
!  TensorProductMultigrid is free software: you can redistribute it and/or
!  modify it under the terms of the GNU Lesser General Public License as
!  published by the Free Software Foundation, either version 3 of the
!  License, or (at your option) any later version.
!  
!  TensorProductMultigrid is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU Lesser General Public License for more details.
!  
!  You should have received a copy of the GNU Lesser General Public License
!  along with TensorProductMultigrid (see files COPYING and COPYING.LESSER).
!  If not, see <http://www.gnu.org/licenses/>.
!
!=== COPYRIGHT AND LICENSE STATEMENT ===


!==================================================================
!
!  Main program for multigrid solver code for Helmholtz/Poisson
!  equation, discretised in the cell centred finite volume scheme
!
!    Eike Mueller, University of Bath, Feb 2012
!
!==================================================================

!==================================================================
! Main program
!==================================================================

module mode_mg_main_mnh

  use discretisation
  use parameters
  use datatypes
  use multigrid
  use conjugategradient
  use solver
  use profiles
  use messages
  use communication
  use timer
#ifndef MNH
  use mpi
#else
  use modd_mpif
#endif

  use mode_mg_read_param
  use mode_mg

contains

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine mg_main_mnh_init(KN,KNZ,PL,PH,PA_K,PB_K,PC_K,PD_K)

    use parameters , only : LMean
    
    implicit none

    integer       , optional , intent (in) :: KN,KNZ
    real(kind=rl) , optional , intent (in) :: PL,PH
    real(kind=rl) , optional , intent (in) :: PA_K(:),PB_K(:),PC_K(:),PD_K(:)


    call mg_init_mnh(KN,KNZ,PL,PH,PA_K,PB_K,PC_K,PD_K)

    ! Initialise ghosts in initial solution, as mg_solve assumes that they
    ! are up-to-date
    call haloswap_mnh(mg_param%n_lev,pproc,xu_fine)

    if (.NOT. PRESENT(KN)) then 
       !
       ! Force some parameter for Idealized StandAlone Newman Solver until convergence
       !
       ! -> Mean must by set to 0 for covergence with Newmann boundaries
       LMean = .true.
       ! -> converge up to 1e-10 in 50 iteration
       solver_param%resreduction =  1.0d-10
       solver_param%maxiter = 50
       
       if (i_am_master_mpi) then
          call flush(STDOUT)
          write(STDOUT,*)
          write(STDOUT,*) "!!!!! WARNING mg_main_mnh_init !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! "
          write(STDOUT,*) "!!!!! FORCED PARAMETER FOR StandAlone Newman Solver until convergence !!!! "
          write(STDOUT,*) "!!!!! LMean                     = " , LMean
          write(STDOUT,*) "!!!!! solver_param%resreduction = " , solver_param%resreduction
          write(STDOUT,*) "!!!!! solver_param%maxiter      = " , solver_param%maxiter
          write(STDOUT,*) "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! "
          write(STDOUT,*)
          call flush(STDOUT)
       end if
    endif
    
  end subroutine mg_main_mnh_init

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine mg_main_initialise_rhs_mnh(KIB,KIE,KIU,KJB,KJE,KJU,KKU,PY)

implicit none

integer       , optional , intent(in) :: KIB,KIE,KIU,KJB,KJE,KJU,KKU
real(kind=rl) , optional , intent(in) :: PY(:,:,:)

 call zt1d_discretisation_init(KIU,KJU,KKU)
 call initialise_rhs_mnh(grid_param,model_param,xb_fine,KIB,KIE,KIU,KJB,KJE,KJU,KKU,PY)
 call haloswap_mnh(mg_param%n_lev,pproc,xb_fine)

end subroutine mg_main_initialise_rhs_mnh
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine mg_main_initialise_u_mnh(KIB,KIE,KIU,KJB,KJE,KJU,KKU,PU)

implicit none

integer       , optional , intent(in) :: KIB,KIE,KIU,KJB,KJE,KJU,KKU
real(kind=rl) , optional , intent(in) :: PU(:,:,:)

 call initialise_u_mnh(grid_param,model_param,xu_fine,KIB,KIE,KIU,KJB,KJE,KJU,KKU,PU)
 call haloswap_mnh(mg_param%n_lev,pproc,xu_fine)

end subroutine mg_main_initialise_u_mnh
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine mg_main_get_u_mnh(KIB,KIE,KIU,KJB,KJE,KJU,KKU,PU)

implicit none

integer       , optional , intent(in)    :: KIB,KIE,KIU,KJB,KJE,KJU,KKU
real(kind=rl) , optional , intent(inout) :: PU(:,:,:)

 call get_u_mnh(grid_param,model_param,xu_fine,KIB,KIE,KIU,KJB,KJE,KJU,KKU,PU)

end subroutine mg_main_get_u_mnh
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine mg_main_mnh_solve()

    implicit none

    ! Solve using multigrid
    call initialise_timer(t_solve,"t_solve")
    call start_timer(t_solve)
    comm_measuretime = .True.

    call mg_solve_mnh(xb_fine,xu_fine,solver_param)

    comm_measuretime = .False.
    call finish_timer(t_solve)

  end subroutine mg_main_mnh_solve

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine mg_main_mnh_finalize()


  end subroutine mg_main_mnh_finalize

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine mg_main_mnh

    implicit none

    integer :: it

    call  mg_main_mnh_init()

    DO it=1,10
       
       call mg_main_initialise_rhs_mnh()
       call mg_main_initialise_u_mnh()

       if (i_am_master_mpi) then
          write(STDOUT,*) '************************ IT=',it,' ***************************'
          call flush(STDOUT)
       end if
       
       call mg_main_mnh_solve()
       
    ENDDO

    call mg_finalize()

  end subroutine mg_main_mnh
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

end module mode_mg_main_mnh

program mg_main_mnh_all
  
  use  mode_mg_main_mnh
  
  call mg_main_mnh()
  
end program mg_main_mnh_all
