module mode_mg

 use datatypes
 use communication
 use discretisation
 use multigrid
 use conjugategradient
 use solver
 use parameters
 use timer
#ifndef MNH
  use mpi
#else
  use modd_mpif
#endif
 use profiles

 use mode_mg_read_param

implicit none

  type(grid_parameters)     :: grid_param
  type(comm_parameters)     :: comm_param
  type(model_parameters)    :: model_param
  type(smoother_parameters) :: smoother_param
  type(mg_parameters)       :: mg_param
  type(cg_parameters)       :: cg_param
  type(solver_parameters)   :: solver_param

  type(scalar3d) :: xu_fine
  type(scalar3d) :: xb_fine
  type(scalar3d) :: xr_fine
#ifdef TESTCONVERGENCE
  type(scalar3d) :: uerror
  real(kind=rl) :: l2error
#endif

 ! --- Name of executable ---
  character(len=256) :: executable

 ! --- Parameter file ---
  character(len=256) :: parameterfile

 ! --- General parameters ---
  logical :: savefields   ! Save fields to disk?

  integer :: i, int_size
  integer :: i_arg
  integer :: ierr

 ! Timers
  type(time) :: t_solve
  type(time) :: t_readparam
  type(time) :: t_initialise
  type(time) :: t_finalise

contains

subroutine mg_init_mnh(KN,KNZ,PL,PH,PA_K,PB_K,PC_K,PD_K)

use MODE_OPENACC_SET_DEVICE
  
implicit none

integer       , optional , intent (in) :: KN,KNZ
real(kind=rl) , optional , intent (in) :: PL,PH
real(kind=rl) , optional , intent (in) :: PA_K(:),PB_K(:),PC_K(:),PD_K(:)

! local var

logical :: gisinit

 ! Initialise MPI ...
  call mpi_initialized(gisinit, ierr )
  if  (.not. gisinit ) then
     call mpi_init(ierr)
  end if
  !
  ! get default device type     
  call MNH_OPENACC_GET_DEVICE_AT_INIT()
  ! ... and pre initialise communication module
  call comm_preinitialise()

    parameterfile="parameters_mg.nam"

  if (i_am_master_mpi) then
    write(STDOUT,*) ''
    write(STDOUT,*) 'Compile time parameters:'
    write(STDOUT,*) ''
#ifdef CARTESIANGEOMETRY
    write(STDOUT,*) '  Geometry: Cartesian'
#else
    write(STDOUT,*) '  Geometry: Spherical'
#endif
#ifdef USELAPACK
    write(STDOUT,*) '  Use Lapack: Yes'
#else
    write(STDOUT,*) '  Use Lapack: No'
#endif
#ifdef OVERLAPCOMMS
    write(STDOUT,*) '  Overlap communications and calculation: Yes'
#else
    write(STDOUT,*) '  Overlap communications and calculation: No'
#endif
    write(STDOUT,*) ''

  end if

  ! Initialise timing module
  call initialise_timing("timing.txt")

  ! Read parameter files
  call initialise_timer(t_readparam,"t_readparam")
  call start_timer(t_readparam)
  if (i_am_master_mpi) then
    write(STDOUT,*) "Reading parameters from file '" // &
                    trim(parameterfile) // "'"
  end if
  call read_general_parameters(parameterfile,savefields)
  call read_solver_parameters(parameterfile,solver_param)
  call read_grid_parameters_mnh(parameterfile,grid_param,KN,KNZ,PL,PH)
  call read_comm_parameters(parameterfile,comm_param)
  call read_model_parameters(parameterfile,model_param)
  call read_smoother_parameters(parameterfile,smoother_param)
  call read_multigrid_parameters(parameterfile,mg_param)
  call read_conjugategradient_parameters(parameterfile,cg_param)
  call finish_timer(t_readparam)

  if (i_am_master_mpi) then
    write(STDOUT,*) ''
  end if

  ! Initialise discretisation module
  call discretisation_initialise_mnh(grid_param,     &
                                 model_param,    &
                                 smoother_param, &
                                 mg_param%n_lev, &
                                 PA_K,PB_K,PC_K,PD_K )

  ! Initialise communication module
  call initialise_timer(t_initialise,"t_initialise")
  call start_timer(t_initialise)
  call comm_initialise(mg_param%n_lev,     &
                       mg_param%lev_split, &
                       grid_param,         &
                       comm_param)

  ! Initialise multigrid
  call mg_initialise(grid_param,       &
                     comm_param,       &
                     model_param,      &
                     smoother_param,   &
                     mg_param,         &
                     cg_param          &
                    )

  call create_scalar3d(MPI_COMM_HORIZ,grid_param,comm_param%halo_size,xu_fine)
  call create_scalar3d(MPI_COMM_HORIZ,grid_param,comm_param%halo_size,xb_fine)
  call create_scalar3d(MPI_COMM_HORIZ,grid_param,comm_param%halo_size,xr_fine)
 
!!$  call  mg_initialise_rhs_mnh()

  call finish_timer(t_initialise)
  if (i_am_master_mpi) then
    write(STDOUT,*) ''
  end if

end subroutine mg_init_mnh

subroutine mg_init()

implicit none

 ! Initialise MPI ...
  call mpi_init(ierr)

  ! ... and pre initialise communication module
  call comm_preinitialise()

    parameterfile="parameters_mg.nam"

  if (i_am_master_mpi) then
    write(STDOUT,*) ''
    write(STDOUT,*) 'Compile time parameters:'
    write(STDOUT,*) ''
#ifdef CARTESIANGEOMETRY
    write(STDOUT,*) '  Geometry: Cartesian'
#else
    write(STDOUT,*) '  Geometry: Spherical'
#endif
#ifdef USELAPACK
    write(STDOUT,*) '  Use Lapack: Yes'
#else
    write(STDOUT,*) '  Use Lapack: No'
#endif
#ifdef OVERLAPCOMMS
    write(STDOUT,*) '  Overlap communications and calculation: Yes'
#else
    write(STDOUT,*) '  Overlap communications and calculation: No'
#endif
    write(STDOUT,*) ''

  end if

  ! Initialise timing module
  call initialise_timing("timing.txt")

  ! Read parameter files
  call initialise_timer(t_readparam,"t_readparam")
  call start_timer(t_readparam)
  if (i_am_master_mpi) then
    write(STDOUT,*) "Reading parameters from file '" // &
                    trim(parameterfile) // "'"
  end if
  call read_general_parameters(parameterfile,savefields)
  call read_solver_parameters(parameterfile,solver_param)
  call read_grid_parameters(parameterfile,grid_param)
  call read_comm_parameters(parameterfile,comm_param)
  call read_model_parameters(parameterfile,model_param)
  call read_smoother_parameters(parameterfile,smoother_param)
  call read_multigrid_parameters(parameterfile,mg_param)
  call read_conjugategradient_parameters(parameterfile,cg_param)
  call finish_timer(t_readparam)

  if (i_am_master_mpi) then
    write(STDOUT,*) ''
  end if

  ! Initialise discretisation module
  call discretisation_initialise(grid_param,     &
                                 model_param,    &
                                 smoother_param, &
                                 mg_param%n_lev )

  ! Initialise communication module
  call initialise_timer(t_initialise,"t_initialise")
  call start_timer(t_initialise)
  call comm_initialise(mg_param%n_lev,     &
                       mg_param%lev_split, &
                       grid_param,         &
                       comm_param)

  ! Initialise multigrid
  call mg_initialise(grid_param,       &
                     comm_param,       &
                     model_param,      &
                     smoother_param,   &
                     mg_param,         &
                     cg_param          &
                    )

  call create_scalar3d(MPI_COMM_HORIZ,grid_param,comm_param%halo_size,xu_fine)
  call create_scalar3d(MPI_COMM_HORIZ,grid_param,comm_param%halo_size,xb_fine)
  call create_scalar3d(MPI_COMM_HORIZ,grid_param,comm_param%halo_size,xr_fine)
  call initialise_rhs(grid_param,model_param,xb_fine)
#ifdef TESTCONVERGENCE
  call create_scalar3d(MPI_COMM_HORIZ,grid_param,comm_param%halo_size,uerror)
  call analytical_solution(grid_param,uerror)
#endif
  call finish_timer(t_initialise)
  if (i_am_master_mpi) then
    write(STDOUT,*) ''
  end if

end subroutine mg_init

subroutine mg_finalize()

implicit none

#ifdef TESTCONVERGENCE
  call daxpy_scalar3d(-1.0_rl,xu_fine,uerror)
  call haloswap(mg_param%n_lev,pproc,uerror)
  l2error = l2norm(uerror)
  if (i_am_master_mpi) then
    write(STDOUT,'("||error|| = ",E20.12," log_2(||error||) = ",E20.12)') &
      l2error, log(l2error)/log(2.0_rl)
  end if
#endif

#ifdef TESTCONVERGENCE
  if (savefields) then
    call save_scalar3d(MPI_COMM_HORIZ,uerror,"error")
  end if
#endif

  ! Save fields to disk
  if (savefields) then
    call haloswap_mnh(mg_param%n_lev,pproc,xu_fine)
    call save_scalar3d(MPI_COMM_HORIZ,xu_fine,"solution")
    call volscale_scalar3d_mnh(xb_fine,1)
    call calculate_residual_mnh(mg_param%n_lev,pproc,xb_fine,xu_fine,xr_fine)
    call volscale_scalar3d_mnh(xb_fine,-1)
    call volscale_scalar3d_mnh(xr_fine,-1)
    call haloswap_mnh(mg_param%n_lev,pproc,xr_fine)
    call save_scalar3d(MPI_COMM_HORIZ,xr_fine,"residual")
    call save_scalar3d(MPI_COMM_HORIZ,xb_fine,"righthand")
  end if

  if (i_am_master_mpi) then
    write(STDOUT,*) ''
  end if

  call discretisation_finalise()

  ! Finalise
  call initialise_timer(t_finalise,"t_finalise")
  call start_timer(t_finalise)
  call mg_finalise()
  call cg_finalise()
  ! Deallocate memory
  call destroy_scalar3d(xu_fine)
  call destroy_scalar3d(xb_fine)
  call destroy_scalar3d(xr_fine)
#ifdef TESTCONVERGENCE
  call destroy_scalar3d(uerror)
#endif


  ! Finalise communications ...
  call comm_finalise(mg_param%n_lev,mg_param%lev_split,grid_param)
  call finish_timer(t_finalise)
  call print_timerinfo("# --- Main timing results ---")
  call print_elapsed(t_readparam,.true.,1.0_rl)
  call print_elapsed(t_initialise,.true.,1.0_rl)
  call print_elapsed(t_solve,.true.,1.0_rl)
  call print_elapsed(t_finalise,.true.,1.0_rl)
  ! Finalise timing
  call finalise_timing()
  ! ... and MPI
  call mpi_finalize(ierr)


end subroutine mg_finalize

end module mode_mg
