!=== COPYRIGHT AND LICENSE STATEMENT ===
!
!  This file is part of the TensorProductMultigrid code.
!  
!  (c) The copyright relating to this work is owned jointly by the
!  Crown, Met Office and NERC [2014]. However, it has been created
!  with the help of the GungHo Consortium, whose members are identified
!  at https://puma.nerc.ac.uk/trac/GungHo/wiki .
!  
!  Main Developer: Eike Mueller
!  
!  TensorProductMultigrid is free software: you can redistribute it and/or
!  modify it under the terms of the GNU Lesser General Public License as
!  published by the Free Software Foundation, either version 3 of the
!  License, or (at your option) any later version.
!  
!  TensorProductMultigrid is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU Lesser General Public License for more details.
!  
!  You should have received a copy of the GNU Lesser General Public License
!  along with TensorProductMultigrid (see files COPYING and COPYING.LESSER).
!  If not, see <http://www.gnu.org/licenses/>.
!
!=== COPYRIGHT AND LICENSE STATEMENT ===

module mode_mg_read_param

contains

!==================================================================
! Read general parameters from namelist file
!==================================================================
subroutine read_general_parameters(filename,savefields_out)
  use parameters
  use communication
  use messages
#ifndef MNH
  use mpi
#else
  use modd_mpif
#endif 
  implicit none
  character(*), intent(in) :: filename
  logical, intent(out) :: savefields_out
  integer, parameter :: file_id = 16
  logical :: savefields
  integer :: ierr
  namelist /parameters_general/ savefields
  if (i_am_master_mpi) then
    open(file_id,file=filename)
    read(file_id,NML=parameters_general)
    close(file_id)
    write(STDOUT,NML=parameters_general)
    write(STDOUT,'("---------------------------------------------- ")')
    write(STDOUT,'("General parameters")')
    write(STDOUT,'("    Save fields = ",L6)') savefields
    write(STDOUT,'("---------------------------------------------- ")')
    write(STDOUT,'("")')
  end if
  call mpi_bcast(savefields,1,MPI_LOGICAL,master_rank,MPI_COMM_WORLD,ierr)
  savefields_out = savefields
end subroutine read_general_parameters

!==================================================================
! Read solver parameters from namelist file
!==================================================================
subroutine read_solver_parameters(filename,solver_param_out)
  use solver
  use parameters
  use communication
  use messages
#ifndef MNH
  use mpi
#else
  use modd_mpif
#endif
  implicit none
  character(*), intent(in) :: filename
  type(solver_parameters), intent(out) :: solver_param_out
  integer :: solvertype
  real(kind=rl) :: resreduction
  integer :: maxiter
  integer, parameter :: file_id = 16
  integer :: ierr
  namelist /parameters_solver/ solvertype,resreduction, maxiter &
                               ,  LUseO , LUseT , LMean
  if (i_am_master_mpi) then
    open(file_id,file=filename)
    read(file_id,NML=parameters_solver)
    close(file_id)
    write(STDOUT,NML=parameters_solver)
    write(STDOUT,'("---------------------------------------------- ")')
    write(STDOUT,'("  LMean = ",L8)') LMean
    write(STDOUT,'("  LUseO = ",L8)') LUseO
    write(STDOUT,'("  LUseT = ",L8)') LUseT
    write(STDOUT,'("Solver parameters ")')
    if (solvertype == SOLVER_RICHARDSON) then
      write(STDOUT,'("    solver       = Richardson")')
    else if (solvertype == SOLVER_CG) then
      write(STDOUT,'("    solver       = CG")')
    else
      call fatalerror("Unknown solver type")
    end if
    write(STDOUT,'("    maxiter      = ",I8)') maxiter
    write(STDOUT,'("    resreduction = ",E15.6)') resreduction
    write(STDOUT,'("---------------------------------------------- ")')
    write(*,'("")')
  end if
  call mpi_bcast(solvertype,1,MPI_INTEGER,master_rank,MPI_COMM_WORLD,ierr)
  call mpi_bcast(maxiter,1,MPI_INTEGER,master_rank,MPI_COMM_WORLD,ierr)
  call mpi_bcast(resreduction,1,MPI_DOUBLE_PRECISION,master_rank,MPI_COMM_WORLD,ierr)
  call mpi_bcast(LMean,1,MPI_LOGICAL,master_rank,MPI_COMM_WORLD,ierr)
  call mpi_bcast(LUseO,1,MPI_LOGICAL,master_rank,MPI_COMM_WORLD,ierr)
  call mpi_bcast(LUseT,1,MPI_LOGICAL,master_rank,MPI_COMM_WORLD,ierr)
  solver_param_out%solvertype = solvertype
  solver_param_out%maxiter = maxiter
  solver_param_out%resreduction = resreduction
end subroutine read_solver_parameters

!==================================================================
! Read grid parameters from namelist file
!==================================================================
subroutine read_grid_parameters_mnh(filename,grid_param,KN,KNZ,PL,PH)
  use parameters
  use datatypes
  use communication
  use messages
#ifndef MNH
  use mpi
#else
  use modd_mpif
#endif 
  implicit none
  character(*), intent(in) :: filename
  type(grid_parameters), intent(out) :: grid_param

  integer       , optional , intent (in) :: KN,KNZ
  real(kind=rl) , optional , intent (in) :: PL,PH

  ! Grid parameters
  integer :: n, nz
  real(kind=rl) :: L, H
  integer :: vertbc
  logical :: graded
  integer, parameter :: file_id = 106
  integer :: ierr
  namelist /parameters_grid/ n, nz, L, H, vertbc, graded
  if (i_am_master_mpi) then
    open(file_id,file=filename)
    read(file_id,NML=parameters_grid)
    close(file_id)

    if (vertbc == VERTBC_DIRICHLET) then
      write(STDOUT,'("    vertbc = DIRICHLET")')
    else if (vertbc == VERTBC_NEUMANN) then
      write(STDOUT,'("    vertbc = NEUMANN")')
    else
      vertbc = -1
    end if
    write(STDOUT,'("    graded =",L3)') graded
    write(STDOUT,'("---------------------------------------------- ")')
    write(STDOUT,'("")')
  end if
  call mpi_bcast(n,1,MPI_INTEGER,master_rank,MPI_COMM_WORLD,ierr)
  call mpi_bcast(nz,1,MPI_INTEGER,master_rank,MPI_COMM_WORLD,ierr)
  call mpi_bcast(L,1,MPI_DOUBLE_PRECISION,master_rank,MPI_COMM_WORLD,ierr)
  call mpi_bcast(H,1,MPI_DOUBLE_PRECISION,master_rank,MPI_COMM_WORLD,ierr)
  call mpi_bcast(vertbc,1,MPI_INTEGER,master_rank,MPI_COMM_WORLD,ierr)
  call mpi_bcast(graded,1,MPI_LOGICAL,master_rank,MPI_COMM_WORLD,ierr)
  IF (PRESENT(KN)) THEN
     n = KN
     nz = KNZ
     L = PL
     H = PH
  END IF
  if (i_am_master_mpi) then
    write(STDOUT,NML=parameters_grid)
    write(STDOUT,'("---------------------------------------------- ")')
    write(STDOUT,'("Grid parameters")')
    write(STDOUT,'("    n      = ",I15)') n
    write(STDOUT,'("    nz     = ",I15)') nz
    write(STDOUT,'("    L      = ",E15.6)') L
    write(STDOUT,'("    H      = ",E15.6)') H
  end if
  grid_param%n = n
  grid_param%nz = nz
  grid_param%L = L
  grid_param%H = H
  grid_param%vertbc = vertbc
  grid_param%graded = graded
  if (vertbc == -1) then
    call fatalerror("vertbc has to be either 1 (Dirichlet) or 2 (Neumann)")
  end if
end subroutine read_grid_parameters_mnh
!==================================================================
! Read grid parameters from namelist file
!==================================================================
subroutine read_grid_parameters(filename,grid_param)
  use parameters
  use datatypes
  use communication
  use messages
#ifndef MNH
  use mpi
#else
  use modd_mpif
#endif
  
  implicit none
  character(*), intent(in) :: filename
  type(grid_parameters), intent(out) :: grid_param
  ! Grid parameters
  integer :: n, nz
  real(kind=rl) :: L, H
  integer :: vertbc
  logical :: graded
  integer, parameter :: file_id = 16
  integer :: ierr
  namelist /parameters_grid/ n, nz, L, H, vertbc, graded
  if (i_am_master_mpi) then
    open(file_id,file=filename)
    read(file_id,NML=parameters_grid)
    close(file_id)
    write(STDOUT,NML=parameters_grid)
    write(STDOUT,'("---------------------------------------------- ")')
    write(STDOUT,'("Grid parameters")')
    write(STDOUT,'("    n      = ",I15)') n
    write(STDOUT,'("    nz     = ",I15)') nz
    write(STDOUT,'("    L      = ",E15.6)') L
    write(STDOUT,'("    H      = ",E15.6)') H
    if (vertbc == VERTBC_DIRICHLET) then
      write(STDOUT,'("    vertbc = DIRICHLET")')
    else if (vertbc == VERTBC_NEUMANN) then
      write(STDOUT,'("    vertbc = NEUMANN")')
    else
      vertbc = -1
    end if
    write(STDOUT,'("    graded =",L3)') graded
    write(STDOUT,'("---------------------------------------------- ")')
    write(STDOUT,'("")')
  end if
  call mpi_bcast(n,1,MPI_INTEGER,master_rank,MPI_COMM_WORLD,ierr)
  call mpi_bcast(nz,1,MPI_INTEGER,master_rank,MPI_COMM_WORLD,ierr)
  call mpi_bcast(L,1,MPI_DOUBLE_PRECISION,master_rank,MPI_COMM_WORLD,ierr)
  call mpi_bcast(H,1,MPI_DOUBLE_PRECISION,master_rank,MPI_COMM_WORLD,ierr)
  call mpi_bcast(vertbc,1,MPI_INTEGER,master_rank,MPI_COMM_WORLD,ierr)
  call mpi_bcast(graded,1,MPI_LOGICAL,master_rank,MPI_COMM_WORLD,ierr)
  grid_param%n = n
  grid_param%nz = nz
  grid_param%L = L
  grid_param%H = H
  grid_param%vertbc = vertbc
  grid_param%graded = graded
  if (vertbc == -1) then
    call fatalerror("vertbc has to be either 1 (Dirichlet) or 2 (Neumann)")
  end if
end subroutine read_grid_parameters
!==================================================================
! Read parallel communication parameters from namelist file
!==================================================================
subroutine read_comm_parameters(filename,comm_param)
  use parameters
  use communication
  use messages
#ifndef MNH
  use mpi
#else
  use modd_mpif
#endif 
  implicit none
  character(*), intent(in) :: filename
  type(comm_parameters), intent(out) :: comm_param
  ! Grid parameters
  integer :: halo_size
  integer, parameter :: file_id = 16
  integer :: ierr
  namelist /parameters_communication/ halo_size
  if (i_am_master_mpi) then
    open(file_id,file=filename)
    read(file_id,NML=parameters_communication)
    close(file_id)
    write(STDOUT,NML=parameters_communication)
    write(STDOUT,'("---------------------------------------------- ")')
    write(STDOUT,'("Communication parameters")')
    write(STDOUT,'("    halosize  = ",I3)') halo_size
    write(STDOUT,'("---------------------------------------------- ")')
    write(STDOUT,'("")')
    if ( (halo_size .ne. 1) ) then
      halo_size = -1
    end if
  end if
  call mpi_bcast(halo_size,1,MPI_INTEGER,master_rank,MPI_COMM_WORLD,ierr)
  comm_param%halo_size = halo_size
  if (halo_size == -1) then
    call fatalerror("Halo size has to be 1.")
  end if
end subroutine read_comm_parameters

!==================================================================
! Read model parameters from namelist file
!==================================================================
subroutine read_model_parameters(filename,model_param)
  use parameters
  use discretisation
  use communication
  use messages
#ifndef MNH
  use mpi
#else
  use modd_mpif
#endif 
  implicit none
  character(*), intent(in) :: filename
  type(model_parameters), intent(out) :: model_param
  real(kind=rl) :: omega2, lambda2, delta
  integer, parameter :: file_id = 16
  integer :: ierr
  namelist /parameters_model/ omega2, lambda2, delta
  if (i_am_master_mpi) then
    open(file_id,file=filename)
    read(file_id,NML=parameters_model)
    close(file_id)
    write(STDOUT,NML=parameters_model)
    write(STDOUT,'("---------------------------------------------- ")')
    write(STDOUT,'("Model parameters")')
    write(STDOUT,'("    omega2  = ",E15.6)') omega2
    write(STDOUT,'("    lambda2 = ",E15.6)') lambda2
    write(STDOUT,'("    delta   = ",E15.6)') delta
    write(STDOUT,'("---------------------------------------------- ")')
    write(STDOUT,'("")')
  end if
  call mpi_bcast(omega2,1,MPI_DOUBLE_PRECISION,master_rank,MPI_COMM_WORLD,ierr)
  call mpi_bcast(lambda2,1,MPI_DOUBLE_PRECISION,master_rank,MPI_COMM_WORLD,ierr)
  call mpi_bcast(delta,1,MPI_DOUBLE_PRECISION,master_rank,MPI_COMM_WORLD,ierr)
  model_param%omega2 = omega2
  model_param%lambda2 = lambda2
  model_param%delta = delta
end subroutine read_model_parameters

!==================================================================
! Read smoother parameters from namelist file
!==================================================================
subroutine read_smoother_parameters(filename,smoother_param)
  use parameters
  use discretisation
  use communication
  use messages
#ifndef MNH
  use mpi
#else
  use modd_mpif
#endif
  implicit none
  character(*), intent(in) :: filename
  type(smoother_parameters), intent(out) :: smoother_param
  integer, parameter :: file_id = 16
  integer :: smoother, ordering
  real(kind=rl) :: rho
  integer :: ierr
  namelist /parameters_smoother/ smoother,           &
                                  ordering,           &
                                  rho
  if (i_am_master_mpi) then
    open(file_id,file=filename)
    read(file_id,NML=parameters_smoother)
    close(file_id)
    write(STDOUT,NML=parameters_smoother)

    write(STDOUT,'("---------------------------------------------- ")')
    write(STDOUT,'("Smoother parameters")')
    ! Smoother
    if (smoother == SMOOTHER_LINE_SOR) then
      write(STDOUT,'("    smoother      = LINE_SOR")')
    else if (smoother == SMOOTHER_LINE_SSOR) then
      write(STDOUT,'("    smoother      = LINE_SSOR")')
    else if (smoother == SMOOTHER_LINE_JAC) then
      write(STDOUT,'("    smoother      = LINE_JACOBI")')
    else
      smoother = -1
    end if

    if (ordering == ORDERING_LEX) then
      write(STDOUT,'("    ordering      = LEX")')
    else if (ordering == ORDERING_RB) then
      write(STDOUT,'("    ordering      = RB")')
    else
      ordering = -1
    end if
    write(STDOUT,'("    rho = ",E15.6)') rho
    write(STDOUT,'("---------------------------------------------- ")')
    write(STDOUT,'("")')
  end if
  call mpi_bcast(smoother,1,MPI_INTEGER,master_rank,MPI_COMM_WORLD,ierr)
  call mpi_bcast(ordering,1,MPI_INTEGER,master_rank,MPI_COMM_WORLD,ierr)
  call mpi_bcast(rho,1,MPI_DOUBLE_PRECISION,master_rank,MPI_COMM_WORLD,ierr)
  smoother_param%smoother = smoother
  smoother_param%ordering = ordering
  smoother_param%rho = rho
  if (smoother == -1) then
    call fatalerror('Unknown smoother.')
  end if
  if (ordering == -1) then
    call fatalerror('Unknown ordering.')
  end if

end subroutine read_smoother_parameters

!==================================================================
! Read multigrid parameters from namelist file
!==================================================================
subroutine read_multigrid_parameters(filename,mg_param)
  use parameters
  use multigrid
  use communication
  use messages
#ifndef MNH
  use mpi
#else
  use modd_mpif
#endif

  use MODE_OPENACC_SET_DEVICE, only : iswitch_cpu_gpu
  
  implicit none
  character(*), intent(in) :: filename
  type(mg_parameters), intent(out) :: mg_param
  integer, parameter :: file_id = 16
  integer :: verbose, n_lev, lev_split, n_presmooth, n_postsmooth,    &
             prolongation, restriction, n_coarsegridsmooth, &
             coarsegridsolver
  integer :: ierr
  namelist /parameters_multigrid/ verbose,            &
                                  n_lev,              &
                                  lev_split,          &
                                  iswitch_cpu_gpu,    &
                                  n_presmooth,        &
                                  n_postsmooth,       &
                                  n_coarsegridsmooth, &
                                  prolongation,       &
                                  restriction,        &
                                  coarsegridsolver
  if (i_am_master_mpi) then
    open(file_id,file=filename)
    read(file_id,NML=parameters_multigrid)
    close(file_id)
    write(STDOUT,NML=parameters_multigrid)
    write(STDOUT,'("---------------------------------------------- ")')
    write(STDOUT,'("Multigrid parameters")')
    write(STDOUT,'("    verbose      = ",I6)') verbose
    write(STDOUT,'("    levels       = ",I3)') n_lev
    write(STDOUT,'("    splitlevel   = ",I3)') lev_split
    !
    ! switch lev_split from the top level to have constant lev_split
    ! when changing n_lev
    !
    lev_split = n_lev - lev_split
    write(STDOUT,'("    splitlevel/sw= ",I3)') lev_split
    write(STDOUT,'(" iswitch_cpu_gpu = ",I3)') iswitch_cpu_gpu
    iswitch_cpu_gpu = n_lev - iswitch_cpu_gpu
    write(STDOUT,'("switch_cpu_gpu/s = ",I3)') iswitch_cpu_gpu
    !
    write(STDOUT,'("    n_presmooth  = ",I6)') n_presmooth
    write(STDOUT,'("    n_postsmooth = ",I6)') n_postsmooth
    if (restriction == REST_CELLAVERAGE) then
      write(STDOUT,'("    restriction   = CELLAVERAGE")')
   else if (restriction == REST_KHALIL) then
      write(STDOUT,'("    restriction   = KHALIL ")')
    else
      restriction = -1
    endif
    if (prolongation == PROL_CONSTANT) then
      write(STDOUT,'("    prolongation  = CONSTANT")')
    else if (prolongation == PROL_TRILINEAR) then
#ifdef PIECEWISELINEAR
      write(STDOUT,'("    prolongation  = TRILINEAR (piecewise linear)")')
#else
      write(STDOUT,'("    prolongation  = TRILINEAR (regression plane)")')
#endif
    else
      prolongation = -1
    endif
    if (coarsegridsolver == COARSEGRIDSOLVER_CG) then
      write(STDOUT,'("    coarse solver = CG")')
    else if (coarsegridsolver == COARSEGRIDSOLVER_SMOOTHER) then
      write(STDOUT,'("    coarse solver = SMOOTHER (",I6," iterations)")') &
        n_coarsegridsmooth
    else
      coarsegridsolver = -1
    end if
    write(STDOUT,'("---------------------------------------------- ")')
    write(*,'("")')

  end if
  call mpi_bcast(verbose,1,MPI_INTEGER,master_rank,MPI_COMM_WORLD,ierr)
  call mpi_bcast(n_lev,1,MPI_INTEGER,master_rank,MPI_COMM_WORLD,ierr)
  call mpi_bcast(lev_split,1,MPI_INTEGER,master_rank,MPI_COMM_WORLD,ierr)
  call mpi_bcast(iswitch_cpu_gpu,1,MPI_INTEGER,master_rank,MPI_COMM_WORLD,ierr)
  call mpi_bcast(n_presmooth,1,MPI_INTEGER,master_rank,MPI_COMM_WORLD,ierr)
  call mpi_bcast(n_postsmooth,1,MPI_INTEGER,master_rank,MPI_COMM_WORLD,ierr)
  call mpi_bcast(n_coarsegridsmooth,1,MPI_INTEGER,master_rank,MPI_COMM_WORLD, &
                 ierr)
  call mpi_bcast(prolongation,1,MPI_INTEGER,master_rank,MPI_COMM_WORLD,ierr)
  call mpi_bcast(restriction,1,MPI_INTEGER,master_rank,MPI_COMM_WORLD,ierr)
  call mpi_bcast(coarsegridsolver,1,MPI_Integer,master_rank,MPI_COMM_WORLD,ierr)
  mg_param%verbose = verbose
  mg_param%n_lev = n_lev
  mg_param%lev_split = lev_split
  mg_param%n_presmooth = n_presmooth
  mg_param%n_postsmooth = n_postsmooth
  mg_param%n_coarsegridsmooth = n_coarsegridsmooth
  mg_param%prolongation = prolongation
  mg_param%restriction = restriction
  mg_param%coarsegridsolver = coarsegridsolver
  if (restriction == -1) then
    call fatalerror('Unknown restriction.')
  end if
  if (prolongation == -1) then
    call fatalerror('Unknown prolongation.')
  end if
  if (coarsegridsolver == -1) then
    call fatalerror('Unknown coarse grid solver.')
  end if
end subroutine read_multigrid_parameters

!==================================================================
! Read conjugate gradient parameters from namelist file
!==================================================================
subroutine read_conjugategradient_parameters(filename,cg_param)
  use parameters
  use communication
  use conjugategradient
  use communication
  use messages
#ifndef MNH
  use mpi
#else
  use modd_mpif
#endif 
  implicit none
  character(*), intent(in) :: filename
  type(cg_parameters), intent(out) :: cg_param
  integer, parameter :: file_id = 16
  integer :: verbose, maxiter, n_prec
  real(kind=rl) :: resreduction
  integer :: ierr
  namelist /parameters_conjugategradient/ verbose,      &
                                          maxiter,      &
                                          resreduction, &
                                          n_prec
  if (i_am_master_mpi) then
    open(file_id,file=filename)
    read(file_id,NML=parameters_conjugategradient)
    close(file_id)
    write(STDOUT,NML=parameters_conjugategradient)
    write(STDOUT,'("---------------------------------------------- ")')
    write(STDOUT,'("Conjugate gradient parameters")')
    write(STDOUT,'("    verbose       = ",I6)') verbose
    write(STDOUT,'("    maxiter       = ",I6)') maxiter
    write(STDOUT,'("    resreduction  = ",E15.6)') resreduction
    write(STDOUT,'("    n_prec        = ",I6)') n_prec
    write(STDOUT,'("---------------------------------------------- ")')
    write(STDOUT,'("")')
  end if
  call mpi_bcast(verbose,1,MPI_INTEGER,master_rank,MPI_COMM_WORLD,ierr)
  call mpi_bcast(maxiter,1,MPI_INTEGER,master_rank,MPI_COMM_WORLD,ierr)
  call mpi_bcast(resreduction,1,MPI_DOUBLE_PRECISION,master_rank,MPI_COMM_WORLD,ierr)
  call mpi_bcast(n_prec,1,MPI_INTEGER,master_rank,MPI_COMM_WORLD,ierr)
  cg_param%verbose = verbose
  cg_param%maxiter = maxiter
  cg_param%resreduction = resreduction
  cg_param%n_prec = n_prec
end subroutine read_conjugategradient_parameters

end module mode_mg_read_param
