!=== COPYRIGHT AND LICENSE STATEMENT ===
!
!  This file is part of the TensorProductMultigrid code.
!  
!  (c) The copyright relating to this work is owned jointly by the
!  Crown, Met Office and NERC [2014]. However, it has been created
!  with the help of the GungHo Consortium, whose members are identified
!  at https://puma.nerc.ac.uk/trac/GungHo/wiki .
!  
!  Main Developer: Eike Mueller
!  
!  TensorProductMultigrid is free software: you can redistribute it and/or
!  modify it under the terms of the GNU Lesser General Public License as
!  published by the Free Software Foundation, either version 3 of the
!  License, or (at your option) any later version.
!  
!  TensorProductMultigrid is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU Lesser General Public License for more details.
!  
!  You should have received a copy of the GNU Lesser General Public License
!  along with TensorProductMultigrid (see files COPYING and COPYING.LESSER).
!  If not, see <http://www.gnu.org/licenses/>.
!
!=== COPYRIGHT AND LICENSE STATEMENT ===


!==================================================================
!
!  Geometric multigrid module for cell centred finite volume
!  discretisation.
!
!    Eike Mueller, University of Bath, Feb 2012
!
!==================================================================
module multigrid

#ifndef MNH
  use mpi
#else
  use modd_mpif
#endif
  use parameters
  use datatypes
  use discretisation
  use messages
  use solver
  use conjugategradient
  use communication
  use timer

  implicit none

public::mg_parameters
public::mg_initialise
public::mg_finalise
public::mg_solve_mnh
public::mg_solve
public::measurehaloswap
public::REST_CELLAVERAGE
public::REST_KHALIL
public::PROL_CONSTANT
public::PROL_TRILINEAR
public::COARSEGRIDSOLVER_SMOOTHER
public::COARSEGRIDSOLVER_CG

private

  ! --- multigrid parameter constants ---
  ! restriction
  integer, parameter :: REST_CELLAVERAGE = 1
  integer, parameter :: REST_KHALIL      = 2
  ! prolongation method
  integer, parameter :: PROL_CONSTANT = 1
  integer, parameter :: PROL_TRILINEAR = 2
  ! Coarse grid solver
  integer, parameter :: COARSEGRIDSOLVER_SMOOTHER = 1
  integer, parameter :: COARSEGRIDSOLVER_CG = 2

  ! --- Multigrid parameters type ---
  type mg_parameters
    ! Verbosity level
    integer :: verbose
    ! Number of MG levels
    integer :: n_lev
    ! First level where data is pulled together
    integer :: lev_split
    ! Number of presmoothing steps
    integer :: n_presmooth
    ! Number of postsmoothing steps
    integer :: n_postsmooth
    ! Number of smoothing steps on coarsest level
    integer :: n_coarsegridsmooth
    ! Prolongation (see PROL_... for allowed values)
    integer :: prolongation
    ! Restriction (see RESTR_... for allowed values)
    integer :: restriction
    ! Smoother (see SMOOTHER_... for allowed values)
    integer :: smoother
    ! Relaxation factor
    real(kind=rl) :: omega
    ! Smoother on coarse grid
    integer :: coarsegridsolver
    ! ordering of grid points for smoother
    integer :: ordering
  end type mg_parameters

! --- Parameters ---
  type(mg_parameters) :: mg_param
  type(model_parameters) :: model_param
  type(smoother_parameters) :: smoother_param
  type(grid_parameters) :: grid_param
  type(comm_parameters) :: comm_param
  type(cg_parameters) :: cg_param


! --- Gridded and scalar data structures ---
  ! Solution vector
  type(scalar3d), allocatable :: xu_mg(:,:)
  ! RHS vector
  type(scalar3d), allocatable :: xb_mg(:,:)
  ! residual
  type(scalar3d), allocatable :: xr_mg(:,:)

! --- Timer ---
  type(time), allocatable, dimension(:,:) :: t_restrict
  type(time), allocatable, dimension(:,:) :: t_prolongate
  type(time), allocatable, dimension(:,:) :: t_residual
  type(time), allocatable, dimension(:,:) :: t_addcorr
  type(time), allocatable, dimension(:,:) :: t_smooth
  type(time), allocatable, dimension(:,:) :: t_coarsesolve
  type(time), allocatable, dimension(:,:) :: t_total

  logical                                 :: Gmg_initialized = .FALSE.

contains

!==================================================================
! Initialise multigrid module, check and print out out parameters
!==================================================================
  subroutine mg_initialise(grid_param_in,     &  ! Grid parameters
                           comm_param_in,     &  ! Comm parameters
                           model_param_in,    &  ! Model parameters
                           smoother_param_in, &  ! Smoother parameters
                           mg_param_in,       &  ! Multigrid parameters
                           cg_param_in        &  ! CG parameters
                           )
    
    use discretisation, only : zt1d_discretisation_allocate3d
    
    implicit none
    
    type(grid_parameters), intent(in)  :: grid_param_in
    type(comm_parameters), intent(in)  :: comm_param_in
    type(model_parameters), intent(in) :: model_param_in
    type(smoother_parameters), intent(in) :: smoother_param_in
    type(mg_parameters), intent(in)    :: mg_param_in
    type(cg_parameters), intent(in)    :: cg_param_in
    real(kind=rl)                      :: L, H
    integer                            :: n, nz, m, nlocal
    logical                            :: reduced_m
    integer                            :: level
    integer                            :: rank, ierr
    integer, dimension(2)              :: p_horiz
    integer, parameter                 :: dim_horiz = 2
    logical                            :: grid_active
    integer                            :: ix_min, ix_max, iy_min, iy_max
    integer                            :: icompx_min, icompx_max, &
                                          icompy_min, icompy_max
    integer                            :: halo_size
    integer                            :: vertbc
    character(len=32)                  :: t_label

    real , dimension(:,:,:) , pointer , contiguous :: zxu_mg_st,zxb_mg_st,zxr_mg_st

    integer :: nlocalx,nlocaly

    Gmg_initialized = .TRUE.
    
    if (i_am_master_mpi) &
      write(STDOUT,*) '*** Initialising multigrid ***'
    ! Check that cell counts are valid
    grid_param = grid_param_in
    comm_param = comm_param_in
    mg_param = mg_param_in
    model_param = model_param_in
    smoother_param = smoother_param_in
    cg_param = cg_param_in
    halo_size = comm_param%halo_size
    vertbc = grid_param%vertbc

    ! Check parameters
    if (grid_param%n < 2**(mg_param%n_lev-1) ) then
      call fatalerror('Number of cells in x-/y- direction has to be at least 2^{n_lev-1}.')
    endif

    if (mod(grid_param%n,2**(mg_param%n_lev-1)) .ne. 0 ) then
      call fatalerror('Number of cells in x-/y- direction is not a multiple of 2^{n_lev-1}.')
    end if
    if (i_am_master_mpi) &
      write(STDOUT,*) ''

    ! Allocate memory for timers
    allocate(t_smooth(mg_param%n_lev,0:pproc))
    allocate(t_total(mg_param%n_lev,0:pproc))
    allocate(t_restrict(mg_param%n_lev,0:pproc))
    allocate(t_residual(mg_param%n_lev,0:pproc))
    allocate(t_prolongate(mg_param%n_lev,0:pproc))
    allocate(t_addcorr(mg_param%n_lev,0:pproc))
    allocate(t_coarsesolve(mg_param%n_lev,0:pproc))

    ! Allocate memory for all levels and initialise fields
    allocate(xu_mg(mg_param%n_lev,0:pproc))
    allocate(xb_mg(mg_param%n_lev,0:pproc))
    allocate(xr_mg(mg_param%n_lev,0:pproc))
    n = grid_param%n
    nlocal = n/(2**pproc)
    nlocalx = n/(2**pprocx)
    nlocaly = n/(2**pprocy)
    nz = grid_param%nz
    L = grid_param%L
    H = grid_param%H
    level = mg_param%n_lev
    m = pprocx ! pproc
    reduced_m = .false.
    !
    call zt1d_discretisation_init(nlocalx,nlocaly,nz)
    !
    ! Work out local processor coordinates (this is necessary to identify
    ! global coordinates)
    call mpi_comm_rank(MPI_COMM_HORIZ,rank,ierr)
    call mpi_cart_coords(MPI_COMM_HORIZ,rank,dim_horiz,p_horiz,ierr)
    if (i_am_master_mpi) then
      write(STDOUT, &
        '(" Global gridsize (x,y,z) (pproc = ",I4," )      : ",I8," x ",I8," x ",I8)') &
        pproc, n, n, nz
    end if
    do while (level > 0)
      if (i_am_master_mpi) &
        write(STDOUT, &
          '(" Local gridsize (x,y,z) on level ",I3," m = ",I4," : ",I8," x ",I8," x ",I8)') &
          level, m, nlocalx, nlocaly, nz
      if ( (nlocalx < 1) .or. (nlocaly < 1) ) then
        call fatalerror('Number of grid points < 1')
      end if

      ! Set sizes of computational grid (take care at boundaries)
      if (p_horiz(1) == 0) then
        icompy_min = 1
      else
        icompy_min = 1 - (halo_size - 1)
      end if

      if (p_horiz(2) == 0) then
        icompx_min = 1
      else
        icompx_min = 1 - (halo_size - 1)
      end if

      if (p_horiz(1) == 2**pprocy-1) then
        icompy_max = nlocaly
      else
        icompy_max = nlocaly + (halo_size - 1)
      end if

      if (p_horiz(2) == 2**pprocx-1) then
        icompx_max = nlocalx
      else
        icompx_max = nlocalx + (halo_size - 1)
      end if

      ! Allocate data
      if (LUseO) then
      allocate(xu_mg(level,m)%s(0:nz+1,                       &
                            1-halo_size:nlocal+halo_size, &
                            1-halo_size:nlocal+halo_size))
      allocate(xb_mg(level,m)%s(0:nz+1,                       &
                            1-halo_size:nlocal+halo_size, &
                            1-halo_size:nlocal+halo_size))
      allocate(xr_mg(level,m)%s(0:nz+1,                       &
                            1-halo_size:nlocal+halo_size, &
                            1-halo_size:nlocal+halo_size))
      xu_mg(level,m)%s(:,:,:) = 0.0_rl
      xb_mg(level,m)%s(:,:,:) = 0.0_rl
      xr_mg(level,m)%s(:,:,:) = 0.0_rl
      endif

      if (LUseT) then
         allocate(zxu_mg_st(1-halo_size:nlocalx+halo_size, &
              1-halo_size:nlocaly+halo_size, &
              0:nz+1))
         !$acc enter data create (zxu_mg_st)
         xu_mg(level,m)%st => zxu_mg_st
         
         allocate(zxb_mg_st(1-halo_size:nlocalx+halo_size, &
              1-halo_size:nlocaly+halo_size, &
              0:nz+1))
         !$acc enter data create (zxb_mg_st)
         xb_mg(level,m)%st => zxb_mg_st
         
         allocate(zxr_mg_st(1-halo_size:nlocalx+halo_size, &
              1-halo_size:nlocaly+halo_size, &
              0:nz+1))
         !$acc enter data create (zxr_mg_st)
         xr_mg(level,m)%st => zxr_mg_st
      
      !$acc kernels
      zxu_mg_st(:,:,:) = 0.0_rl
      zxb_mg_st(:,:,:) = 0.0_rl
      zxr_mg_st(:,:,:) = 0.0_rl
      !$acc end kernels
      endif

      ! NB: 1st coordinate is in the y-direction of the processor grid,
      ! second coordinate is in the x-direction (see comments in
      ! communication module)
      iy_min = (p_horiz(1)/2**(pproc-m))*nlocaly+1
      iy_max = (p_horiz(1)/2**(pproc-m)+1)*nlocaly
      ix_min = p_horiz(2)/2**(pproc-m)*nlocalx+1
      ix_max = (p_horiz(2)/2**(pproc-m)+1)*nlocalx
      ! Set grid parameters and local data ranges
      ! Note that only n (and possibly nz) change as we
      ! move down the levels
      xu_mg(level,m)%grid_param%L = L
      xu_mg(level,m)%grid_param%H = H
      xu_mg(level,m)%grid_param%n = n
      xu_mg(level,m)%grid_param%nz = nz
      xu_mg(level,m)%grid_param%vertbc = vertbc
      xu_mg(level,m)%ix_min = ix_min
      xu_mg(level,m)%ix_max = ix_max
      xu_mg(level,m)%iy_min = iy_min
      xu_mg(level,m)%iy_max = iy_max
      xu_mg(level,m)%icompx_min = icompx_min
      xu_mg(level,m)%icompx_max = icompx_max
      xu_mg(level,m)%icompy_min = icompy_min
      xu_mg(level,m)%icompy_max = icompy_max
      xu_mg(level,m)%halo_size = halo_size

      xb_mg(level,m)%grid_param%L = L
      xb_mg(level,m)%grid_param%H = H
      xb_mg(level,m)%grid_param%n = n
      xb_mg(level,m)%grid_param%nz = nz
      xb_mg(level,m)%grid_param%vertbc = vertbc
      xb_mg(level,m)%ix_min = ix_min
      xb_mg(level,m)%ix_max = ix_max
      xb_mg(level,m)%iy_min = iy_min
      xb_mg(level,m)%iy_max = iy_max
      xb_mg(level,m)%icompx_min = icompx_min
      xb_mg(level,m)%icompx_max = icompx_max
      xb_mg(level,m)%icompy_min = icompy_min
      xb_mg(level,m)%icompy_max = icompy_max
      xb_mg(level,m)%halo_size = halo_size

      xr_mg(level,m)%grid_param%L = L
      xr_mg(level,m)%grid_param%H = H
      xr_mg(level,m)%grid_param%n = n
      xr_mg(level,m)%grid_param%nz = nz
      xr_mg(level,m)%grid_param%vertbc = vertbc
      xr_mg(level,m)%ix_min = ix_min
      xr_mg(level,m)%ix_max = ix_max
      xr_mg(level,m)%iy_min = iy_min
      xr_mg(level,m)%iy_max = iy_max
      xr_mg(level,m)%icompx_min = icompx_min
      xr_mg(level,m)%icompx_max = icompx_max
      xr_mg(level,m)%icompy_min = icompy_min
      xr_mg(level,m)%icompy_max = icompy_max
      xr_mg(level,m)%halo_size = halo_size

      ! Are these grids active?
      if ( (m == pproc) .or. &
           ( (mod(p_horiz(1),2**(pprocx-m)) == 0) .and. &
             (mod(p_horiz(2),2**(pprocx-m)) == 0) ) ) then
        grid_active = .true.
      else
        grid_active = .false.
      end if
      xu_mg(level,m)%isactive = grid_active
      xb_mg(level,m)%isactive = grid_active
      xr_mg(level,m)%isactive = grid_active
      write(t_label,'("t_total(",I3,",",I3,")")') level, m
      call initialise_timer(t_total(level,m),t_label)
      write(t_label,'("t_smooth(",I3,",",I3,")")') level, m
      call initialise_timer(t_smooth(level,m),t_label)
      write(t_label,'("t_restrict(",I3,",",I3,")")') level, m
      call initialise_timer(t_restrict(level,m),t_label)
      write(t_label,'("t_residual(",I3,",",I3,")")') level, m
      call initialise_timer(t_residual(level,m),t_label)
      write(t_label,'("t_prolongate(",I3,",",I3,")")') level, m
      call initialise_timer(t_prolongate(level,m),t_label)
      write(t_label,'("t_addcorrection(",I3,",",I3,")")') level, m
      call initialise_timer(t_addcorr(level,m),t_label)
      write(t_label,'("t_coarsegridsolver(",I3,",",I3,")")') level, m
      call initialise_timer(t_coarsesolve(level,m),t_label)

      ! If we are below L_split, split data
      if ( (level .le. mg_param%lev_split) .and. &
           (m > 0) .and. (.not. reduced_m) ) then
        reduced_m = .true.
        m = m-1
        nlocal = 2*nlocal
        nlocalx = 2*nlocalx
        nlocaly = 2*nlocaly
        cycle
      end if
      reduced_m = .false.
      level = level-1
      n = n/2
      nlocal = nlocal/2
      nlocalx = nlocalx/2
      nlocaly = nlocaly/2
    end do
    if (i_am_master_mpi) &
      write(STDOUT,*) ''
    call cg_initialise(cg_param)
  end subroutine mg_initialise

!==================================================================
! Finalise, free memory for all data structures
!==================================================================
  subroutine mg_finalise()
    implicit none
    integer :: level, m
    logical :: reduced_m
    character(len=80) :: s
    integer :: ierr

    If ( .not. Gmg_initialized ) return
    
    if (i_am_master_mpi) &
      write(STDOUT,*) '*** Finalising multigrid ***'
    ! Deallocate memory
    level = mg_param%n_lev
    m = pproc
    reduced_m = .false.
    !if (mg_param%verbose >= 0 )
    call print_timerinfo("--- V-cycle timing results ---")
    do while (level > 0)
       if (mg_param%verbose >= 10 ) then
          write(s,'("level = ",I3,", m = ",I3)') level,m
          call print_timerinfo(s)
          call print_elapsed(t_smooth(level,m),.True.,1.0_rl)
          call print_elapsed(t_restrict(level,m),.True.,1.0_rl)
          call print_elapsed(t_prolongate(level,m),.True.,1.0_rl)
          call print_elapsed(t_residual(level,m),.True.,1.0_rl)
          call print_elapsed(t_addcorr(level,m),.True.,1.0_rl)
          call print_elapsed(t_coarsesolve(level,m),.True.,1.0_rl)
          call print_elapsed(t_total(level,m),.True.,1.0_rl)
       endif
      if (LUseO) then
      deallocate(xu_mg(level,m)%s)
      deallocate(xb_mg(level,m)%s)
      deallocate(xr_mg(level,m)%s)
      endif

      if (LUseT) then
         !$acc exit data delete(xu_mg(level,m)%st)  
         deallocate(xu_mg(level,m)%st)
         !
         !$acc exit data delete(xb_mg(level,m)%st)
         deallocate(xb_mg(level,m)%st)
         !
         !$acc exit data delete(xr_mg(level,m)%st)
         deallocate(xr_mg(level,m)%st)
      endif

      ! If we are below L_split, split data
      if ( (level .le. mg_param%lev_split) .and. &
           (m > 0) .and. (.not. reduced_m) ) then
        reduced_m = .true.
        m = m-1
        cycle
      end if
      reduced_m = .false.
      level = level-1
    end do
    deallocate(xu_mg)
    deallocate(xb_mg)
    deallocate(xr_mg)
    deallocate(t_total)
    deallocate(t_smooth)
    deallocate(t_restrict)
    deallocate(t_prolongate)
    deallocate(t_residual)
    deallocate(t_addcorr)
    deallocate(t_coarsesolve)
    !if ( (i_am_master_mpi) ) write(STDOUT,'("")')
    if ( (i_am_master_mpi) .and. (mg_param%verbose >= 10) ) write(STDOUT,'("")')    
  end subroutine mg_finalise

!==================================================================
! Restrict from fine -> coarse
!==================================================================
  subroutine restrict_mnh(phifine,phicoarse)
    implicit none
    type(scalar3d), intent(in)    :: phifine
    type(scalar3d), intent(inout) :: phicoarse
    ! local var
    integer :: ix,iy,iz
    integer :: ix_min, ix_max, iy_min, iy_max, n ,nz
    real(kind=rl) :: xn,xs,xw,xe

    real , dimension(:,:,:) , pointer , contiguous :: zphifine_st , zphicoarse_st

    n      = phicoarse%grid_param%n
    ix_min = phicoarse%icompx_min
    ix_max = phicoarse%icompx_max
    iy_min = phicoarse%icompy_min
    iy_max = phicoarse%icompy_max
    nz     = phicoarse%grid_param%nz
    ! three dimensional cell average
    if (mg_param%restriction == REST_CELLAVERAGE) then
      ! Do not coarsen in z-direction
      if (LUseO) then 
      do ix=ix_min,ix_max
        do iy=iy_min,iy_max
          do iz=1,nz
            phicoarse%s(iz,iy,ix) =  &
              phifine%s(iz  ,2*iy  ,2*ix  ) + &
              phifine%s(iz  ,2*iy-1,2*ix  ) + &
              phifine%s(iz  ,2*iy  ,2*ix-1) + &
              phifine%s(iz  ,2*iy-1,2*ix-1)
          end do
        end do
      end do
      endif
      if (LUseT) then 
         zphifine_st => phifine%st
         zphicoarse_st => phicoarse%st
         !$acc kernels present_cr(zphifine_st,zphicoarse_st)
         !$mnh_do_concurrent( ix=ix_min:ix_max, iy=iy_min:iy_max, iz=1:nz )
              zphicoarse_st(ix,iy,iz) =  &
                       zphifine_st(2*ix  ,2*iy  ,iz) + &
                       zphifine_st(2*ix  ,2*iy-1,iz) + &
                       zphifine_st(2*ix-1,2*iy  ,iz) + &
                       zphifine_st(2*ix-1,2*iy-1,iz)
         !$mnh_end_do()            
         !$acc end kernels
      endif

   elseif(mg_param%restriction == REST_KHALIL) then
      if (LUseO) then 
         do ix=ix_min,ix_max
            xw=1.0
            xe=1.0
            if (ix==1) xw=0.0
            if (ix==n) xe=0.0
            do iy=iy_min,iy_max
               xs=1.0
               xn=1.0
               if (iy==1) xs=0.0
               if (iy==n) xn=0.0
               do iz=1,nz
                  phicoarse%s(iz,iy,ix) = 0.25_rl *         ( &
                       phifine%s(iz,2*iy+1,2*ix-1) * xn        + &
                       phifine%s(iz,2*iy+1,2*ix  ) * xn        + &
                       phifine%s(iz,2*iy  ,2*ix-2) * xw        + &
                       phifine%s(iz,2*iy  ,2*ix-1) * (4-xw-xn) + &
                       phifine%s(iz,2*iy  ,2*ix  ) * (4-xe-xn) + &
                       phifine%s(iz,2*iy  ,2*ix+1) * xe        + &
                       phifine%s(iz,2*iy-1,2*ix-2) * xw        + &
                       phifine%s(iz,2*iy-1,2*ix-1) * (4-xw-xs) + &
                       phifine%s(iz,2*iy-1,2*ix  ) * (4-xe-xs) + &
                       phifine%s(iz,2*iy-2,2*ix-1) * xs        + &
                       phifine%s(iz,2*iy-2,2*ix  ) * xs          &     
                       & )   
               end do
            end do
         end do
      end if
      if (LUseT) then 
         do iz=1,nz
            do iy=iy_min,iy_max
               xs=1.0
               xn=1.0
               if (iy==1) xs=0.0
               if (iy==n) xn=0.0
               do ix=ix_min,ix_max
                  xw=1.0
                  xe=1.0
                  if (ix==1) xw=0.0
                  if (ix==n) xe=0.0
                  phicoarse%st(ix,iy,iz) = 0.25_rl *         ( &
                       phifine%s(2*ix-1,2*iy+1,iz) * xn        + &
                       phifine%s(2*ix  ,2*iy+1,iz) * xn        + &
                       phifine%s(2*ix-2,2*iy  ,iz) * xw        + &
                       phifine%s(2*ix-1,2*iy  ,iz) * (4-xw-xn) + &
                       phifine%s(2*ix  ,2*iy  ,iz) * (4-xe-xn) + &
                       phifine%s(2*ix+1,2*iy  ,iz) * xe        + &
                       phifine%s(2*ix-2,2*iy-1,iz) * xw        + &
                       phifine%s(2*ix-1,2*iy-1,iz) * (4-xw-xs) + &
                       phifine%s(2*ix  ,2*iy-1,iz) * (4-xe-xs) + &
                       phifine%s(2*ix-1,2*iy-2,iz) * xs        + &
                       phifine%s(2*ix  ,2*iy-2,iz) * xs          &     
                       & )                
               end do
            end do
         end do
      end if
      
    end if
  end subroutine restrict_mnh
!==================================================================
! Restrict from fine -> coarse
!==================================================================
  subroutine restrict(phifine,phicoarse)
    implicit none
    type(scalar3d), intent(in)    :: phifine
    type(scalar3d), intent(inout) :: phicoarse
    integer :: ix,iy,iz
    integer :: ix_min, ix_max, iy_min, iy_max , nz

    ix_min = phicoarse%icompx_min
    ix_max = phicoarse%icompx_max
    iy_min = phicoarse%icompy_min
    iy_max = phicoarse%icompy_max
    nz     = phicoarse%grid_param%nz
    
    ! three dimensional cell average
    if (mg_param%restriction == REST_CELLAVERAGE) then
      ! Do not coarsen in z-direction
      if (LUseO) then
         do ix=ix_min,ix_max
            do iy=iy_min,iy_max
               do iz=1,nz
                  phicoarse%s(iz,iy,ix) =  &
                       phifine%s(iz  ,2*iy  ,2*ix  ) + &
                       phifine%s(iz  ,2*iy-1,2*ix  ) + &
                       phifine%s(iz  ,2*iy  ,2*ix-1) + &
                       phifine%s(iz  ,2*iy-1,2*ix-1)
               end do
            end do
         end do
      end if
      if (LUseT) then 
         do iz=1,nz
            do iy=iy_min,iy_max
               do ix=ix_min,ix_max
                  phicoarse%st(ix,iy,iz) =  &
                       phifine%st(2*ix  ,2*iy  ,iz) + &
                       phifine%st(2*ix  ,2*iy-1,iz) + &
                       phifine%st(2*ix-1,2*iy  ,iz) + &
                       phifine%st(2*ix-1,2*iy-1,iz)
               end do
            end do
         end do
      endif
   end if
 end subroutine restrict
!==================================================================
! Prolongate from coarse -> fine
! level, m is the correspong to the fine grid level
!==================================================================
  subroutine prolongate_mnh(level,m,phicoarse,phifine)
    implicit none
    integer, intent(in) :: level
    integer, intent(in) :: m
    type(scalar3d), intent(in) :: phicoarse
    type(scalar3d), intent(inout) :: phifine
    real(kind=rl) :: tmp
    integer :: nlocal
    integer, dimension(5) :: ixmin, ixmax, iymin, iymax
    integer :: n, nz
    integer :: ix, iy, iz
    integer :: dix, diy, diz
    real(kind=rl) :: rhox, rhoy, rhoz
    real(kind=rl) :: rho_i, sigma_j, h, c1, c2
    logical :: overlap_comms
    integer, dimension(4) :: send_requests, recv_requests
    integer, dimension(4) :: send_requestsT, recv_requestsT
    integer :: ierr
    integer :: iblock

    integer :: nlocalx,nlocaly

    ! Needed for interpolation matrix
#ifdef PIECEWISELINEAR
#else
    real(kind=rl) :: dx(4,3), A(3,3), dx_fine(4,2)
    integer :: i,j,k
    real(kind=rl) :: dxu(2), grad(2)
    dx(1,3) = 1.0_rl
    dx(2,3) = 1.0_rl
    dx(3,3) = 1.0_rl
    dx(4,3) = 1.0_rl
#endif

    nlocal = phicoarse%ix_max-phicoarse%ix_min+1
    nlocalx = phicoarse%ix_max-phicoarse%ix_min+1
    nlocaly = phicoarse%iy_max-phicoarse%iy_min+1
    n = phicoarse%grid_param%n
    nz = phicoarse%grid_param%nz

#ifdef OVERLAPCOMMS
    overlap_comms = (nlocal > 2)
#else
    overlap_comms = .false.
#endif
    ! Block 1 (N)
    ixmin(1) = 1
    ixmax(1) = nlocalx
    iymin(1) = 1
    iymax(1) = 1
    ! Block 2 (S)
    ixmin(2) = 1
    ixmax(2) = nlocalx
    iymin(2) = nlocaly
    iymax(2) = nlocaly
    ! Block 3 (W)
    ixmin(3) = 1
    ixmax(3) = 1
    iymin(3) = 2
    iymax(3) = nlocaly-1
    ! Block 4 (E)
    ixmin(4) = nlocalx
    ixmax(4) = nlocalx
    iymin(4) = 2
    iymax(4) = nlocaly-1
    ! Block 5 (INTERIOR)
    if (overlap_comms) then
      ixmin(5) = 2
      ixmax(5) = nlocalx-1
      iymin(5) = 2
      iymax(5) = nlocaly-1
    else
      ! If there are no interior cells, do not overlap
      ! communications and calculations, just loop over interior cells
      ixmin(5) = 1
      ixmax(5) = nlocalx
      iymin(5) = 1
      iymax(5) = nlocaly
    end if

    ! *** Constant prolongation or (tri-) linear prolongation ***
    if ( (mg_param%prolongation == PROL_CONSTANT) .or. &
         (mg_param%prolongation == PROL_TRILINEAR) ) then
      if (overlap_comms) then
        ! Loop over cells next to boundary (iblock = 1,...,4)
        do iblock = 1, 4
          if (mg_param%prolongation == PROL_CONSTANT) then
            call loop_over_grid_constant_mnh(iblock)
          end if
          if (mg_param%prolongation == PROL_TRILINEAR) then
            call loop_over_grid_linear_mnh(iblock)
          end if
        end do
        ! Initiate halo exchange
        call ihaloswap_mnh(level,m,phifine,send_requests,recv_requests,send_requestsT,recv_requestsT)
      end if
      ! Loop over INTERIOR cells
      iblock = 5
      if (mg_param%prolongation == PROL_CONSTANT) then
        call loop_over_grid_constant_mnh(iblock)
      end if
      if (mg_param%prolongation == PROL_TRILINEAR) then
        call loop_over_grid_linear_mnh(iblock)
      end if
      if (overlap_comms) then
        if (m > 0) then
          if (LUseO) call mpi_waitall(4,recv_requests, MPI_STATUSES_IGNORE, ierr)
          if (LUseO) call mpi_waitall(4,send_requests, MPI_STATUSES_IGNORE, ierr)
          if (LUseT) call mpi_waitall(4,recv_requestsT, MPI_STATUSES_IGNORE, ierr)
          if (LUseT) call mpi_waitall(4,send_requestsT, MPI_STATUSES_IGNORE, ierr)
        end if
      else
        call haloswap_mnh(level,m,phifine)
      end if
    else
      call fatalerror("Unsupported prolongation.")
    end if

    contains

    !------------------------------------------------------------------
    ! The actual loops over the grid for the individual blocks,
    ! when overlapping calculation and communication
    !------------------------------------------------------------------

    !------------------------------------------------------------------
    ! (1) Constant interpolation
    !------------------------------------------------------------------
    subroutine loop_over_grid_constant_mnh(iblock)
      implicit none
      integer, intent(in) :: iblock
      integer :: ix,iy,iz, nz

      nz = phicoarse%grid_param%nz
      
      if (LUseO) then
         do ix=ixmin(iblock),ixmax(iblock)
            do iy=iymin(iblock),iymax(iblock)
               do dix = -1,0
                  do diy = -1,0
                     do iz=1,nz
                        phifine%s(iz,2*iy+diy,2*ix+dix) = phicoarse%s(iz,iy,ix)
                     end do
                  end do
               end do
            end do
         end do
      end if
     if (LUseT) then
         do ix=ixmin(iblock),ixmax(iblock)
            do iy=iymin(iblock),iymax(iblock)
               do dix = -1,0
                  do diy = -1,0
                     do iz=1,nz
                        phifine%st(2*ix+dix,2*iy+diy,iz) = phicoarse%st(ix,iy,iz)
                     end do
                  end do
               end do
            end do
         end do
      end if

    end subroutine loop_over_grid_constant_mnh

    !------------------------------------------------------------------
    ! (2) Linear interpolation
    !------------------------------------------------------------------
    subroutine loop_over_grid_linear_mnh(iblock)
      implicit none
      integer, intent(in) :: iblock
      !local var
      integer :: ix,iy,iz

      real , dimension(:,:,:) , pointer , contiguous :: zphifine_st , zphicoarse_st

      integer :: nz
      integer :: ix_min, ix_max, iy_min, iy_max

      nz = phicoarse%grid_param%nz
      ix_min = ixmin(iblock)
      ix_max = ixmax(iblock)
      iy_min = iymin(iblock)
      iy_max = iymax(iblock)
      
      ! optimisation for newman MNH case : all coef constant
      rhox = 0.25_rl
      rhoy = 0.25_rl

      if (LUseO) then     
         do ix=ixmin(iblock),ixmax(iblock)
            do iy=iymin(iblock),iymax(iblock)
               ! Piecewise linear interpolation
               do iz=1,nz
                  do dix = -1,0
                     do diy = -1,0
                        phifine%s(iz,2*iy+diy,2*ix+dix) =      &
                             phicoarse%s(iz,iy,ix) +                &
                             rhox*(phicoarse%s(iz,iy,ix+(2*dix+1))  &
                             - phicoarse%s(iz,iy,ix)) +       &
                             rhoy*(phicoarse%s(iz,iy+(2*diy+1),ix)  &
                             - phicoarse%s(iz,iy,ix))
                     end do
                  end do
               end do
            end do
         end do
      end if
      if (LUseT) then  
         ! Piecewise linear interpolation
         
         zphifine_st => phifine%st
         zphicoarse_st => phicoarse%st

         !!!!!$acc kernels
#ifdef MNH_COMPILER_NVHPC         
         !!!!!$acc loop independent  collapse(5)
#endif         
         !$acc parallel loop collapse(5)
         do diy = -1,0
            do dix = -1,0
               !!!!DO CONCURRENT (ix=ix_min:ix_max, iy=iy_min:iy_max, iz=1:nz)
               do iz=1,nz ; do iy=iy_min,iy_max; do ix=ix_min,ix_max;
                  zphifine_st(2*ix+dix,2*iy+diy,iz) =      &
                       zphicoarse_st(ix,iy,iz) +                &
                       rhox*(zphicoarse_st(ix+(2*dix+1),iy,iz)  &
                       - zphicoarse_st(ix,iy,iz)) +       &
                       rhoy*(zphicoarse_st(ix,iy+(2*diy+1),iz)  &
                       - zphicoarse_st(ix,iy,iz))                  
               end do; end do; end do
               end do
            end do
         !!!!!$acc end kernels
      end if

    end subroutine loop_over_grid_linear_mnh

  end subroutine prolongate_mnh
!==================================================================
! Prolongate from coarse -> fine
! level, m is the correspong to the fine grid level
!==================================================================
  subroutine prolongate(level,m,phicoarse,phifine)
    implicit none
    integer, intent(in) :: level
    integer, intent(in) :: m
    type(scalar3d), intent(in) :: phicoarse
    type(scalar3d), intent(inout) :: phifine
    real(kind=rl) :: tmp
    integer :: nlocal
    integer, dimension(5) :: ixmin, ixmax, iymin, iymax
    integer :: n, nz
    integer :: ix, iy, iz
    integer :: dix, diy, diz
    real(kind=rl) :: rhox, rhoy, rhoz
    real(kind=rl) :: rho_i, sigma_j, h, c1, c2
    logical :: overlap_comms
    integer, dimension(4) :: send_requests, recv_requests
    integer :: ierr
    integer :: iblock

    ! Needed for interpolation matrix
#ifdef PIECEWISELINEAR
#else
    real(kind=rl) :: dx(4,3), A(3,3), dx_fine(4,2)
    integer :: i,j,k
    real(kind=rl) :: dxu(2), grad(2)
    dx(1,3) = 1.0_rl
    dx(2,3) = 1.0_rl
    dx(3,3) = 1.0_rl
    dx(4,3) = 1.0_rl
#endif

    nlocal = phicoarse%ix_max-phicoarse%ix_min+1
    n = phicoarse%grid_param%n
    nz = phicoarse%grid_param%nz

#ifdef OVERLAPCOMMS
    overlap_comms = (nlocal > 2)
#else
    overlap_comms = .false.
#endif
    ! Block 1 (N)
    ixmin(1) = 1
    ixmax(1) = nlocal
    iymin(1) = 1
    iymax(1) = 1
    ! Block 2 (S)
    ixmin(2) = 1
    ixmax(2) = nlocal
    iymin(2) = nlocal
    iymax(2) = nlocal
    ! Block 3 (W)
    ixmin(3) = 1
    ixmax(3) = 1
    iymin(3) = 2
    iymax(3) = nlocal-1
    ! Block 4 (E)
    ixmin(4) = nlocal
    ixmax(4) = nlocal
    iymin(4) = 2
    iymax(4) = nlocal-1
    ! Block 5 (INTERIOR)
    if (overlap_comms) then
      ixmin(5) = 2
      ixmax(5) = nlocal-1
      iymin(5) = 2
      iymax(5) = nlocal-1
    else
      ! If there are no interior cells, do not overlap
      ! communications and calculations, just loop over interior cells
      ixmin(5) = 1
      ixmax(5) = nlocal
      iymin(5) = 1
      iymax(5) = nlocal
    end if

    ! *** Constant prolongation or (tri-) linear prolongation ***
    if ( (mg_param%prolongation == PROL_CONSTANT) .or. &
         (mg_param%prolongation == PROL_TRILINEAR) ) then
      if (overlap_comms) then
        ! Loop over cells next to boundary (iblock = 1,...,4)
        do iblock = 1, 4
          if (mg_param%prolongation == PROL_CONSTANT) then
            call loop_over_grid_constant(iblock)
          end if
          if (mg_param%prolongation == PROL_TRILINEAR) then
            call loop_over_grid_linear(iblock)
          end if
        end do
        ! Initiate halo exchange
        call ihaloswap(level,m,phifine,send_requests,recv_requests)
      end if
      ! Loop over INTERIOR cells
      iblock = 5
      if (mg_param%prolongation == PROL_CONSTANT) then
        call loop_over_grid_constant(iblock)
      end if
      if (mg_param%prolongation == PROL_TRILINEAR) then
        call loop_over_grid_linear(iblock)
      end if
      if (overlap_comms) then
        if (m > 0) then
          call mpi_waitall(4,recv_requests, MPI_STATUSES_IGNORE, ierr)
        end if
      else
        call haloswap(level,m,phifine)
      end if
    else
      call fatalerror("Unsupported prolongation.")
    end if

    contains

    !------------------------------------------------------------------
    ! The actual loops over the grid for the individual blocks,
    ! when overlapping calculation and communication
    !------------------------------------------------------------------

    !------------------------------------------------------------------
    ! (1) Constant interpolation
    !------------------------------------------------------------------
    subroutine loop_over_grid_constant(iblock)
      implicit none
      integer, intent(in) :: iblock
      integer :: ix,iy,iz
      do ix=ixmin(iblock),ixmax(iblock)
        do iy=iymin(iblock),iymax(iblock)
          do dix = -1,0
            do diy = -1,0
              do iz=1,nz
                phifine%s(iz,2*iy+diy,2*ix+dix) = phicoarse%s(iz,iy,ix)
              end do
            end do
          end do
        end do
      end do
    end subroutine loop_over_grid_constant

    !------------------------------------------------------------------
    ! (2) Linear interpolation
    !------------------------------------------------------------------
    subroutine loop_over_grid_linear(iblock)
      implicit none
      integer, intent(in) :: iblock
      integer :: ix,iy,iz
      do ix=ixmin(iblock),ixmax(iblock)
        do iy=iymin(iblock),iymax(iblock)
#ifdef PIECEWISELINEAR
          ! Piecewise linear interpolation
          do iz=1,nz
            do dix = -1,0
              do diy = -1,0
                if ( (ix+(2*dix+1)+phicoarse%ix_min-1  < 1 ) .or. &
                     (ix+(2*dix+1)+phicoarse%ix_min-1  > n ) ) then
                  rhox = 0.5_rl
                else
                  rhox = 0.25_rl
                end if
                if ( (iy+(2*diy+1)+phicoarse%iy_min-1  < 1 ) .or. &
                     (iy+(2*diy+1)+phicoarse%iy_min-1  > n ) ) then
                  rhoy = 0.5_rl
                else
                  rhoy = 0.25_rl
                end if
                 phifine%s(iz,2*iy+diy,2*ix+dix) =      &
                  phicoarse%s(iz,iy,ix) +                &
                  rhox*(phicoarse%s(iz,iy,ix+(2*dix+1))  &
                        - phicoarse%s(iz,iy,ix)) +       &
                  rhoy*(phicoarse%s(iz,iy+(2*diy+1),ix)  &
                        - phicoarse%s(iz,iy,ix))
              end do
            end do
          end do
#else
          ! Fit a plane through the four neighbours of each
          ! coarse grid point. Use the gradient of this plane and
          ! the value of the field on the coarse grid point for
          ! the linear interpolation
          ! Calculate the displacement vectors
#ifdef CARTESIANGEOMETRY
          ! (ix-1, iy)
          dx(1,1) = -1.0_rl
          dx(1,2) =  0.0_rl
          ! (ix+1, iy)
          dx(2,1) = +1.0_rl
          dx(2,2) =  0.0_rl
          ! (ix, iy-1)
          dx(3,1) =  0.0_rl
          dx(3,2) = -1.0_rl
          ! (ix, iy+1)
          dx(4,1) =  0.0_rl
          dx(4,2) = +1.0_rl
#else
          rho_i = 2.0_rl*(ix+(phicoarse%ix_min-1)-0.5_rl)/n-1.0_rl
          sigma_j = 2.0_rl*(iy+(phicoarse%iy_min-1)-0.5_rl)/n-1.0_rl
          if (abs(rho_i**2+sigma_j**2) > 1.0E-12) then
            c1 = (1.0_rl+rho_i**2+sigma_j**2)/sqrt(rho_i**2+sigma_j**2)
            c2 = sqrt(1.0_rl+rho_i**2+sigma_j**2)/sqrt(rho_i**2+sigma_j**2)
          else
            rho_i = 1.0_rl
            sigma_j = 1.0_rl
            c1 = sqrt(0.5_rl)
            c2 = sqrt(0.5_rl)
          end if
          ! (ix-1, iy)
          dx(1,1) = -c1*rho_i
          dx(1,2) = +c2*sigma_j
          ! (ix+1, iy)
          dx(2,1) = +c1*rho_i
          dx(2,2) = -c2*sigma_j
          ! (ix, iy-1)
          dx(3,1) = -c1*sigma_j
          dx(3,2) = -c2*rho_i
          ! (ix, iy+1)
          dx(4,1) = +c1*sigma_j
          dx(4,2) = +c2*rho_i
          dx_fine(1,1) = 0.25_rl*(dx(1,1)+dx(3,1))
          dx_fine(1,2) = 0.25_rl*(dx(1,2)+dx(3,2))
          dx_fine(2,1) = 0.25_rl*(dx(2,1)+dx(3,1))
          dx_fine(2,2) = 0.25_rl*(dx(2,2)+dx(3,2))
          dx_fine(3,1) = 0.25_rl*(dx(1,1)+dx(4,1))
          dx_fine(3,2) = 0.25_rl*(dx(1,2)+dx(4,2))
          dx_fine(4,1) = 0.25_rl*(dx(2,1)+dx(4,1))
          dx_fine(4,2) = 0.25_rl*(dx(2,2)+dx(4,2))
#endif
          ! Boundaries
          if (ix-1+phicoarse%ix_min-1  < 1 ) then
            dx(1,1) = 0.5_rl*dx(1,1)
            dx(1,2) = 0.5_rl*dx(1,2)
          end if
          if (ix+1+phicoarse%ix_min-1  > n ) then
            dx(2,1) = 0.5_rl*dx(2,1)
            dx(2,2) = 0.5_rl*dx(2,2)
          end if
          if (iy-1+phicoarse%iy_min-1  < 1 ) then
            dx(3,1) = 0.5_rl*dx(3,1)
            dx(3,2) = 0.5_rl*dx(3,2)
          end if
          if (iy+1+phicoarse%iy_min-1  > n ) then
            dx(4,1) = 0.5_rl*dx(4,1)
            dx(4,2) = 0.5_rl*dx(4,2)
          end if
          ! Calculate matrix used for least squares linear fit
          A(:,:) = 0.0_rl
          do i = 1,4
            do j=1,3
              do k=1,3
                A(j,k) = A(j,k) + dx(i,j)*dx(i,k)
              end do
            end do
          end do
          ! invert A
          call invertA(A)
          do iz=1,nz
            ! Calculate gradient on each level
            dxu(1:2) = dx(1,1:2)*phicoarse%s(iz,iy  ,ix-1) &
                     + dx(2,1:2)*phicoarse%s(iz,iy  ,ix+1) &
                     + dx(3,1:2)*phicoarse%s(iz,iy-1,ix  ) &
                     + dx(4,1:2)*phicoarse%s(iz,iy+1,ix  )
            grad(:) = 0.0_rl
            do j=1,2
              do k=1,2
                grad(j) = grad(j) + A(j,k)*dxu(k)
              end do
            end do
            ! Use the gradient and the field value in the centre of
            ! the coarse grid cell to interpolate to the fine grid
            ! cells
#ifdef CARTESIANGEOMETRY
            do dix=-1,0
              do diy=-1,0
                phifine%s(iz,2*iy+diy,2*ix+dix) =       &
                  phicoarse%s(iz,iy,ix)                 &
                  + 0.25_rl*( grad(1)*(2.0*dix+1)    &
                            + grad(2)*(2.0*diy+1))
              end do
            end do
#else
            phifine%s(iz,2*iy-1, 2*ix-1) = phicoarse%s(iz,iy,ix) + &
                                         ( grad(1)*dx_fine(1,1)  + &
                                           grad(2)*dx_fine(1,2) )
            phifine%s(iz,2*iy-1, 2*ix  ) = phicoarse%s(iz,iy,ix) + &
                                         ( grad(1)*dx_fine(2,1)  + &
                                           grad(2)*dx_fine(2,2) )
            phifine%s(iz,2*iy  , 2*ix-1) = phicoarse%s(iz,iy,ix) + &
                                         ( grad(1)*dx_fine(3,1)  + &
                                           grad(2)*dx_fine(3,2) )
            phifine%s(iz,2*iy  , 2*ix  ) = phicoarse%s(iz,iy,ix) + &
                                         ( grad(1)*dx_fine(4,1)  + &
                                           grad(2)*dx_fine(4,2) )
#endif
          end do
#endif
        end do
      end do
    end subroutine loop_over_grid_linear

  end subroutine prolongate
  !------------------------------------------------------------------
  ! Invert the 3x3 matrix A either using LaPack or the explicit
  ! formula
  !------------------------------------------------------------------
  subroutine invertA(A)
    implicit none
    real(kind=rl), intent(inout), dimension(3,3) :: A
    real(kind=rl), dimension(3,3) :: Anew
    real(kind=rl) :: invdetA
    integer :: ipiv(3), info
    real(kind=rl) :: work(3)
#ifdef USELAPACK
    call DGETRF( 3, 3, A, 3, ipiv, info )
    call DGETRI( 3, A, 3, ipiv, work, 3, info )
#else
    invdetA = 1.0_rl / ( A(1,1) * (A(3,3)*A(2,2) - A(3,2)*A(2,3)) &
                       - A(2,1) * (A(3,3)*A(1,2) - A(3,2)*A(1,3)) &
                       + A(3,1) * (A(2,3)*A(1,2) - A(2,2)*A(1,3)) )
    Anew(1,1) =     A(3,3)*A(2,2) - A(3,2)*A(2,3)
    Anew(1,2) = - ( A(3,3)*A(1,2) - A(3,2)*A(1,3) )
    Anew(1,3) =     A(2,3)*A(1,2) - A(2,2)*A(1,3)
    Anew(2,1) = - ( A(3,3)*A(2,1) - A(3,1)*A(2,3) )
    Anew(2,2) =     A(3,3)*A(1,1) - A(3,1)*A(1,3)
    Anew(2,3) = - ( A(2,3)*A(1,1) - A(2,1)*A(1,3) )
    Anew(3,1) =     A(3,2)*A(2,1) - A(3,1)*A(2,2)
    Anew(3,2) = - ( A(3,2)*A(1,1) - A(3,1)*A(1,2) )
    Anew(3,3) =     A(2,2)*A(1,1) - A(2,1)*A(1,2)
    A(:,:) = invdetA*Anew(:,:)
#endif
  end subroutine invertA

!==================================================================
! Multigrid V-cycle
!==================================================================
  recursive subroutine mg_vcycle_mnh(b,u,r,finelevel,splitlevel,level,m)
#ifdef MNH_MPPDB
    USE MODE_MPPDB
#endif
    use MODE_OPENACC_SET_DEVICE
    implicit none
    integer, intent(in)                                     :: finelevel
    type(scalar3d), intent(inout), dimension(finelevel,0:pproc) :: b
    type(scalar3d), intent(inout), dimension(finelevel,0:pproc) :: u
    type(scalar3d), intent(inout), dimension(finelevel,0:pproc) :: r
    integer, intent(in)                                     :: splitlevel
    integer, intent(in)                                     :: level
    integer, intent(in)                                     :: m
    !local var
    integer                                                 :: n_gridpoints
    integer                                                 :: nlocalx, nlocaly
    integer                                                 :: halo_size

    real , dimension(:,:,:) , pointer , contiguous ::  zu_level_1_m_st,zu_level_1_m_1_st

    character(len=3) :: ylevel

    nlocalx = u(level,m)%ix_max-u(level,m)%ix_min+1
    nlocaly = u(level,m)%iy_max-u(level,m)%iy_min+1
    halo_size = u(level,m)%halo_size
    n_gridpoints = (nlocalx+2*halo_size) &
                 * (nlocaly+2*halo_size) &
                 * (u(level,m)%grid_param%nz+2)
#ifdef MNH_MPPDB
    if ( mppdb_initialized ) then
       WRITE(ylevel, '( I3.3 )' ) level
     call Mppdb_check3D_real_mg( b(level,m)%st," mg_vcycle_mnh beg:"//ylevel//":b"       )
     call Mppdb_check3D_real_mg( u(level,m)%st," mg_vcycle_mnh beg:"//ylevel//":u"       )
     call Mppdb_check3D_real_mg( r(level,m)%st," mg_vcycle_mnh beg:"//ylevel//":r"       )    
  end if
#endif

    if (level > 1) then
      ! Perform n_presmooth smoothing steps
      call start_timer(t_smooth(level,m))
      call start_timer(t_total(level,m))
#ifdef MG_DEBUG
      if (i_am_master_mpi)  write(STDOUT,*) ' ====== begin  smooth_mnh DIRECTION_FORWARD ====== level=',level
      call print_scalaprod2(level   , m  , b(level  ,m)  , "Befor smooth_mnh b(level  ,m  )=" )
      call print_scalaprod2(level   , m  , u(level  ,m)  , "Befor smooth_mnh u(level  ,m  )=" )            
#endif
      call smooth_mnh(level,m,mg_param%n_presmooth, &
                  DIRECTION_FORWARD, &
                  b(level,m),u(level,m))
#ifdef MG_DEBUG
      call print_scalaprod2(level   , m  , b(level  ,m)  , "After smooth_mnh b(level  ,m  )=" )
      call print_scalaprod2(level   , m  , u(level  ,m)  , "After smooth_mnh u(level  ,m  )=" )                  
#endif
      call finish_timer(t_total(level,m))
      call finish_timer(t_smooth(level,m))
      ! Calculate residual
      call start_timer(t_residual(level,m))
      call start_timer(t_total(level,m))
      call calculate_residual_mnh(level,m,b(level,m),u(level,m),r(level,m))
      call finish_timer(t_total(level,m))
      call finish_timer(t_residual(level,m))
      ! Restrict residual
      call start_timer(t_restrict(level,m))
      call start_timer(t_total(level,m))
#ifdef MG_DEBUG
      if (i_am_master_mpi)  write(STDOUT,*) ' ====== begin  restrict_mnh ====== level=',level
      call print_scalaprod2(level-1 , m  , b(level-1,m)  , "Befor restrict_mnh b(level-1,m  )=" )
      call print_scalaprod2(level   , m  , r(level  ,m)  , "Befor restrict_mnh r(level  ,m  )=" )      
#endif
      call restrict_mnh(r(level,m),b(level-1,m))
#ifdef MG_DEBUG
      call print_scalaprod2(level-1 , m  , b(level-1,m)  , "After restrict_mnh b(level-1,m  )=" )
      call print_scalaprod2(level   , m  , r(level  ,m)  , "After restrict_mnh r(level  ,m  )=" )
#endif
      call finish_timer(t_total(level,m))
      call finish_timer(t_restrict(level,m))
      if ( ((level-1) .le. splitlevel) .and. (m > 0) ) then
        ! Collect data on coarser level
        call start_timer(t_total(level,m))
        call collect(level-1,m,b(level-1,m),b(level-1,m-1))
#ifdef MG_DEBUG
        if (i_am_master_mpi)  write(STDOUT,*) ' ====== after collect ====== level-1=',level-1
        call print_scalaprod2(level-1, m  , b(level-1,m)  , "After collect b(level-1,m  )=" )
        call print_scalaprod2(level-1, m-1, b(level-1,m-1), "After collect b(level-1,m-1)=" )
#endif
        call finish_timer(t_total(level,m))
        ! Set initial solution on coarser level to zero (incl. halos!)
        if (LUseO) u(level-1,m-1)%s(:,:,:) = 0.0_rl
        if (LUseT) then
           zu_level_1_m_1_st => u(level-1,m-1)%st(:,:,:)
           !$acc kernels
           zu_level_1_m_1_st(:,:,:) = 0.0_rl
           !$acc end kernels
        end if
        ! solve on coarser grid
        call mg_vcycle_mnh(b,u,r,finelevel,splitlevel,level-1,m-1)
        ! Distribute data on coarser grid
        call start_timer(t_total(level,m))
        call distribute(level-1,m,u(level-1,m-1),u(level-1,m))
#ifdef MG_DEBUG
        if (i_am_master_mpi)  write(STDOUT,*) ' ====== after distribute ====== level-1=',level-1
        call print_scalaprod2(level-1, m  , u(level-1,m)  , "After distribute u(level-1,m  )=" )
        call print_scalaprod2(level-1, m-1, u(level-1,m-1), "After distribute u(level-1,m-1)=" )       
#endif
        call finish_timer(t_total(level,m))
      else
        ! Set initial solution on coarser level to zero (incl. halos!)
        if (LUseO) u(level-1,m)%s(:,:,:) = 0.0_rl
        if (LUseT) then
           zu_level_1_m_st =>  u(level-1,m)%st(:,:,:)
           !$acc kernels 
           zu_level_1_m_st(:,:,:) = 0.0_rl
           !$acc end kernels
        end if
        ! solve on coarser grid
        ! switch from GPU to CPU if level == iswitch_cpu_gpu 
        if (level .EQ. iswitch_cpu_gpu ) then
           !print*,' enter mg_vcycle_mnh level=', iswitch_cpu_gpu
           !call MNH_OPENACC_GET_DEVICE()
           call MNH_OPENACC_SET_DEVICE_HOST()
           !call MNH_OPENACC_GET_DEVICE()
        end if
        call mg_vcycle_mnh(b,u,r,finelevel,splitlevel,level-1,m)       
        if (level .EQ. iswitch_cpu_gpu) then
           !print*,' exit mg_vcycle_mnh level=', iswitch_cpu_gpu
           !call MNH_OPENACC_GET_DEVICE()
           call MNH_OPENACC_SET_DEVICE_DEFAULT()
           !call MNH_OPENACC_GET_DEVICE()
        end if
      end if
      ! Prolongate error
      call start_timer(t_prolongate(level,m))
      call start_timer(t_total(level,m))

#ifdef MG_DEBUG
      if (i_am_master_mpi)  write(STDOUT,*) ' ====== begin haloswap_mnh_mnh ====== level-1=',level-1
      call print_scalaprod2(level-1 , m  , u(level-1,m)  , "Befor haloswap_mnh u(level-1,m  )=" )
#endif
      call haloswap_mnh(level-1,m,u(level-1,m))
#ifdef MG_DEBUG
      call print_scalaprod2(level-1 , m  , u(level-1,m)  , "After haloswap_mnh u(level-1,m  )=" )

      if (i_am_master_mpi)  write(STDOUT,*) ' ====== begin boundary_mnh_mnh ====== level-1=',level-1
      call print_scalaprod2(level-1 , m  , u(level-1,m)  , "Befor boundary_mnh u(level-1,m  )=" )
#endif
      call boundary_mnh(u(level-1,m))
#ifdef MG_DEBUG
      call print_scalaprod2(level-1 , m  , u(level-1,m)  , "After boundary_mnh u(level-1,m  )=" )
      
      if (i_am_master_mpi)  write(STDOUT,*) ' ====== begin prolongate_mnh ====== level=',level
      call print_scalaprod2(level-1 , m  , u(level-1,m)  , "Befor prolongate_mnh u(level-1,m  )=" )
      call print_scalaprod2(level   , m  , r(level  ,m)  , "Befor prolongate_mnh r(level  ,m  )=" )
#endif
      call prolongate_mnh(level,m,u(level-1,m),r(level,m))
#ifdef MG_DEBUG
      call print_scalaprod2(level-1 , m  , u(level-1,m)  , "After prolongate_mnh u(level-1,m  )=" )
      call print_scalaprod2(level   , m  , r(level  ,m)  , "After prolongate_mnh r(level  ,m  )=" )
#endif
      call finish_timer(t_total(level,m))
      call finish_timer(t_prolongate(level,m))
      ! Add error to fine grid solution
      call start_timer(t_addcorr(level,m))
      call start_timer(t_total(level,m))
      if (LUseO) call daxpy(n_gridpoints,1.0_rl,r(level,m)%s,1,u(level,m)%s,1)
      if (LUseT) call daxpy(n_gridpoints,1.0_rl,r(level,m)%st,1,u(level,m)%st,1)
      call finish_timer(t_total(level,m))
      call finish_timer(t_addcorr(level,m))
      ! Perform n_postsmooth smoothing steps
      call start_timer(t_smooth(level,m))
      call start_timer(t_total(level,m))
      call smooth_mnh(level,m, &
                  mg_param%n_postsmooth, &
                  DIRECTION_BACKWARD, &
                  b(level,m),u(level,m))
      call finish_timer(t_total(level,m))
      call finish_timer(t_smooth(level,m))
    else
      call start_timer(t_coarsesolve(level,m))
      call start_timer(t_total(level,m))
      if (mg_param%coarsegridsolver == COARSEGRIDSOLVER_CG) then
        call cg_solve_mnh(level,m,b(level,m),u(level,m))
        call print_scalaprod2(level, m , u(level,m), "After cg_solve_mnh u=(level,m)" )
      else if (mg_param%coarsegridsolver == COARSEGRIDSOLVER_SMOOTHER) then
        ! Smooth on coarsest level
        call smooth_mnh(level,m, &
                    mg_param%n_coarsegridsmooth,     &
                    DIRECTION_FORWARD, &
                    b(level,m),u(level,m))
      end if
      call finish_timer(t_total(level,m))
      call finish_timer(t_coarsesolve(level,m))
    end if
#ifdef MNH_MPPDB
    if ( mppdb_initialized ) then
       WRITE(ylevel, '( I3.3 )' ) level
     call Mppdb_check3D_real_mg( b(level,m)%st," mg_vcycle_mnh end:"//ylevel//":b"       )
     call Mppdb_check3D_real_mg( u(level,m)%st," mg_vcycle_mnh end:"//ylevel//":u"       )
     call Mppdb_check3D_real_mg( r(level,m)%st," mg_vcycle_mnh end:"//ylevel//":r"       )    
  end if
#endif
    
  end subroutine mg_vcycle_mnh
!==================================================================
! Multigrid V-cycle
!==================================================================
  recursive subroutine mg_vcycle(b,u,r,finelevel,splitlevel,level,m)
    implicit none
    integer, intent(in)                                     :: finelevel
    type(scalar3d), intent(inout), dimension(finelevel,0:pproc) :: b
    type(scalar3d), intent(inout), dimension(finelevel,0:pproc) :: u
    type(scalar3d), intent(inout), dimension(finelevel,0:pproc) :: r
    integer, intent(in)                                     :: splitlevel
    integer, intent(in)                                     :: level
    integer, intent(in)                                     :: m
    integer                                                 :: n_gridpoints
    integer                                                 :: nlocalx, nlocaly
    integer                                                 :: halo_size

    nlocalx = u(level,m)%ix_max-u(level,m)%ix_min+1
    nlocaly = u(level,m)%iy_max-u(level,m)%iy_min+1
    halo_size = u(level,m)%halo_size
    n_gridpoints = (nlocalx+2*halo_size) &
                 * (nlocaly+2*halo_size) &
                 * (u(level,m)%grid_param%nz+2)

    if (level > 1) then
      ! Perform n_presmooth smoothing steps
      call start_timer(t_smooth(level,m))
      call start_timer(t_total(level,m))
      call smooth(level,m,mg_param%n_presmooth, &
                  DIRECTION_FORWARD, &
                  b(level,m),u(level,m))
      call finish_timer(t_total(level,m))
      call finish_timer(t_smooth(level,m))
      ! Calculate residual
      call start_timer(t_residual(level,m))
      call start_timer(t_total(level,m))
      call calculate_residual(level,m,b(level,m),u(level,m),r(level,m))
      call finish_timer(t_total(level,m))
      call finish_timer(t_residual(level,m))
      ! Restrict residual
      call start_timer(t_restrict(level,m))
      call start_timer(t_total(level,m))
      call restrict(r(level,m),b(level-1,m))
      call finish_timer(t_total(level,m))
      call finish_timer(t_restrict(level,m))
      if ( ((level-1) .le. splitlevel) .and. (m >= 0) ) then
        ! Collect data on coarser level
        call start_timer(t_total(level,m))
        call collect(level-1,m,b(level-1,m),b(level-1,m-1))
        call haloswap_mnh(level-1,m-1,b(level-1,m-1))
        call finish_timer(t_total(level,m))
        ! Set initial solution on coarser level to zero (incl. halos!)
        if (LUseO) u(level-1,m-1)%s(:,:,:) = 0.0_rl
        if (LUseT) u(level-1,m-1)%st(:,:,:) = 0.0_rl
        ! solve on coarser grid
        call mg_vcycle(b,u,r,finelevel,splitlevel,level-1,m-1)
        ! Distribute data on coarser grid
        call start_timer(t_total(level,m))
        call distribute(level-1,m,u(level-1,m-1),u(level-1,m))
        call haloswap(level-1,m,u(level-1,m))
        call finish_timer(t_total(level,m))
      else
        ! Set initial solution on coarser level to zero (incl. halos!)
        if (LUseO) u(level-1,m)%s(:,:,:) = 0.0_rl
        if (LUseT) u(level-1,m)%st(:,:,:) = 0.0_rl
        ! solve on coarser grid
        call mg_vcycle(b,u,r,finelevel,splitlevel,level-1,m)
      end if
      ! Prolongate error
      call start_timer(t_prolongate(level,m))
      call start_timer(t_total(level,m))
      call prolongate(level,m,u(level-1,m),r(level,m))
      call finish_timer(t_total(level,m))
      call finish_timer(t_prolongate(level,m))
      ! Add error to fine grid solution
      call start_timer(t_addcorr(level,m))
      call start_timer(t_total(level,m))
      if (LUseO) call daxpy(n_gridpoints,1.0_rl,r(level,m)%s,1,u(level,m)%s,1)
      if (LUseT) call daxpy(n_gridpoints,1.0_rl,r(level,m)%st,1,u(level,m)%st,1)
      call finish_timer(t_total(level,m))
      call finish_timer(t_addcorr(level,m))
      ! Perform n_postsmooth smoothing steps
      call start_timer(t_smooth(level,m))
      call start_timer(t_total(level,m))
      call smooth(level,m, &
                  mg_param%n_postsmooth, &
                  DIRECTION_BACKWARD, &
                  b(level,m),u(level,m))
      call finish_timer(t_total(level,m))
      call finish_timer(t_smooth(level,m))
    else
      call start_timer(t_coarsesolve(level,m))
      call start_timer(t_total(level,m))
      if (mg_param%coarsegridsolver == COARSEGRIDSOLVER_CG) then
        call cg_solve(level,m,b(level,m),u(level,m))
      else if (mg_param%coarsegridsolver == COARSEGRIDSOLVER_SMOOTHER) then
        ! Smooth on coarsest level
        call smooth(level,m, &
                    mg_param%n_coarsegridsmooth,     &
                    DIRECTION_FORWARD, &
                    b(level,m),u(level,m))
      end if
      call finish_timer(t_total(level,m))
      call finish_timer(t_coarsesolve(level,m))
    end if

  end subroutine mg_vcycle

!==================================================================
! Test halo exchanges
!==================================================================
  recursive subroutine mg_vcyclehaloswaponly(b,u,r,finelevel,splitlevel,level,m)
    implicit none
    integer, intent(in)                                     :: finelevel
    type(scalar3d), intent(inout), dimension(finelevel,0:pproc) :: b
    type(scalar3d), intent(inout), dimension(finelevel,0:pproc) :: u
    type(scalar3d), intent(inout), dimension(finelevel,0:pproc) :: r
    integer, intent(in)                                     :: splitlevel
    integer, intent(in)                                     :: level
    integer, intent(in)                                     :: m
    integer, parameter :: nhaloswap = 100
    integer :: i
    integer :: ierr

    if (level > 1) then
      call mpi_barrier(MPI_COMM_HORIZ,ierr)
      do i=1,nhaloswap
        call haloswap(level,m,u(level,m))
      end do
      call mpi_barrier(MPI_COMM_HORIZ,ierr)
      if ( ((level-1) .le. splitlevel) .and. (m >= 0) ) then
        call mpi_barrier(MPI_COMM_HORIZ,ierr)
        do i=1,nhaloswap
          call haloswap(level-1,m,u(level-1,m))
        end do
        call mpi_barrier(MPI_COMM_HORIZ,ierr)
        call mg_vcyclehaloswaponly(b,u,r,finelevel,splitlevel,level-1,m-1)
      else
        call mg_vcyclehaloswaponly(b,u,r,finelevel,splitlevel,level-1,m)
      end if
    else
      ! Haloswap on coarsest level
      call mpi_barrier(MPI_COMM_HORIZ,ierr)
      do i=1,nhaloswap
        call haloswap(level,m,u(level,m))
      end do
      call mpi_barrier(MPI_COMM_HORIZ,ierr)
    end if

  end subroutine mg_vcyclehaloswaponly

!==================================================================
! Multigrid solve
! Assumes that ghosts in initial solution are up-to-date
!==================================================================
  subroutine mg_solve_mnh(bRHS,usolution,solver_param)
#ifdef MNH_MPPDB
    USE MODE_MPPDB
#endif    
    implicit none
    type(scalar3d), intent(in)      :: bRHS
    type(scalar3d), intent(inout)   :: usolution
    type(solver_parameters), intent(in) :: solver_param
    integer :: solvertype
    real(kind=rl)  :: resreduction
    integer :: maxiter
    integer :: n_gridpoints
    integer :: iter, level, finelevel, splitlevel
    real(kind=rl) :: res_old, res_new, res_initial , mean_initial , norm_initial
    logical :: solverconverged = .false.
    integer :: nlocalx, nlocaly
    integer :: halo_size
    type(time) :: t_prec, t_res, t_apply, t_l2norm, t_scalprod, t_mainloop
    type(scalar3d) :: pp
    type(scalar3d) :: q
    type(scalar3d) :: z_one 
    real(kind=rl) :: alpha, beta, pq, rz, rz_old
    integer :: ierr
    real(kind=rl) , pointer , contiguous , dimension(:,:,:) :: z_one_st
    integer :: icompx_max, icompy_max, nz

    solvertype = solver_param%solvertype
    resreduction = solver_param%resreduction
    maxiter = solver_param%maxiter
    nlocalx = usolution%ix_max - usolution%ix_min+1
    nlocaly = usolution%iy_max - usolution%iy_min+1
    halo_size = usolution%halo_size

    level = mg_param%n_lev
    finelevel = level
    splitlevel = mg_param%lev_split

    ! Initialise timers
    call initialise_timer(t_prec,"t_prec")
    call initialise_timer(t_apply,"t_apply")
    call initialise_timer(t_l2norm,"t_l2norm")
    call initialise_timer(t_scalprod,"t_scalarprod")
    call initialise_timer(t_res,"t_residual")
    call initialise_timer(t_mainloop,"t_mainloop")

    n_gridpoints = (nlocalx+2*halo_size) &
                 * (nlocaly+2*halo_size) &
                 * (usolution%grid_param%nz+2)
    !
    ! Init 1 vector = z_one
    call create_scalar3d(MPI_COMM_HORIZ,bRHS%grid_param,halo_size,z_one)
    !
    icompx_max = z_one%icompx_max
    icompy_max = z_one%icompy_max
    nz         = z_one%grid_param%nz
    !
    if (LUseO) then
       z_one%s(:,:,:) = 0.0_rl
       z_one%s(1:nz,1:icompy_max,1:icompx_max) = 1.0_rl
    end if
    if (LUseT) then       
       z_one_st => z_one%st
       !$acc kernels
       z_one_st(:,:,:) = 0.0_rl
       z_one_st(1:icompx_max,1:icompy_max,1:nz) = 1.0_rl
       !$acc end kernels
    end if
    !   Mean / Norm of B
    call scalarprod_mnh(pproc,z_one,z_one, mean_initial ) 
    call scalarprod_mnh(pproc,z_one,bRHS, mean_initial ) 
    norm_initial =  l2norm_mnh(bRHS,.true.)
    mean_initial =  mean_initial / (( nz ) * ( z_one%grid_param%n )**2)
    norm_initial =  mean_initial / norm_initial
    if (LMean) then
       ! b <- b -mean_initial * z_one
       if (LUseO) call daxpy(n_gridpoints,-mean_initial,z_one%s,1,bRHS%s,1)
       if (LUseT) call daxpy(n_gridpoints,-mean_initial,z_one%st,1,bRHS%st,1)
       call scalarprod_mnh(pproc,z_one,bRHS, mean_initial ) 
    endif
    !
    ! Copy b to b(1)
    ! Copy usolution to u(1)
    if (LUseO) call dcopy(n_gridpoints,bRHS%s,1,xb_mg(level,pproc)%s,1)
    if (LUseT) call dcopy(n_gridpoints,bRHS%st,1,xb_mg(level,pproc)%st,1)
    if (LUseO) call dcopy(n_gridpoints,usolution%s,1,xu_mg(level,pproc)%s,1)
    if (LUseT) call dcopy(n_gridpoints,usolution%st,1,xu_mg(level,pproc)%st,1)

! Scale with volume of grid cells
    call volscale_scalar3d_mnh(xb_mg(level,pproc),1)
    call scalarprod_mnh(pproc,z_one,xb_mg(level,pproc), mean_initial )

    call start_timer(t_res)
    call calculate_residual_mnh(level, pproc, &
                            xb_mg(level,pproc),xu_mg(level,pproc),xr_mg(level,pproc))
    call finish_timer(t_res)
    call start_timer(t_l2norm)
    res_initial = l2norm_mnh(xr_mg(level,pproc),.true.)
    call finish_timer(t_l2norm)
    res_old = res_initial
    if (mg_param%verbose > 0) then
      if (i_am_master_mpi) then
        write(STDOUT,'(" *** Multigrid solver ***")')
        write(STDOUT,'(" <MG> Initial residual : ",E23.10)') res_initial
      end if
    end if
    if (mg_param%verbose > 0) then
      if (i_am_master_mpi) then
        write(STDOUT,'(" <MG>   iter :   residual         rho")')
        write(STDOUT,'(" <MG> --------------------------------")')
      end if
    end if

    call mpi_barrier(MPI_COMM_WORLD,ierr)
    call start_timer(t_mainloop)
    if (solvertype == SOLVER_CG) then
      ! NB: b(level,pproc) will be used as r in the following
      call create_scalar3d(MPI_COMM_HORIZ,bRHS%grid_param,halo_size,pp)
      call create_scalar3d(MPI_COMM_HORIZ,bRHS%grid_param,halo_size,q)
      ! Apply preconditioner: Calculate p = M^{-1} r
      ! (1) copy b <- r
      if (LUseO) call dcopy(n_gridpoints,xr_mg(level,pproc)%s,1,xb_mg(level,pproc)%s,1)
      if (LUseT) call dcopy(n_gridpoints,xr_mg(level,pproc)%st,1,xb_mg(level,pproc)%st,1)
      ! (2) set u <- 0
      if (LUseO) xu_mg(level,pproc)%s(:,:,:) = 0.0_rl
      if (LUseT) xu_mg(level,pproc)%st(:,:,:) = 0.0_rl
      ! (3) Call MG Vcycle
      call start_timer(t_prec)
      call mg_vcycle_mnh(xb_mg,xu_mg,xr_mg,finelevel,splitlevel,level,pproc)
      call finish_timer(t_prec)
      ! (4) copy pp <- u (=solution from MG Vcycle)
      if (LUseO) call dcopy(n_gridpoints,xu_mg(level,pproc)%s,1,pp%s,1)
      if (LUseT) call dcopy(n_gridpoints,xu_mg(level,pproc)%st,1,pp%st,1)
      ! Calculate rz_old = <pp,b>
      call start_timer(t_scalprod)
      call scalarprod_mnh(pproc,pp,xb_mg(level,pproc),rz_old)
      call finish_timer(t_scalprod)
      do iter = 1, maxiter
        ! Apply matrix q <- A.pp
        call start_timer(t_apply)
        call apply_mnh(pp,q)
        call finish_timer(t_apply)
        ! Calculate pq <- <pp,q>
        call start_timer(t_scalprod)
        call scalarprod_mnh(pproc,pp,q,pq)
        call finish_timer(t_scalprod)
        alpha = rz_old/pq
        ! x <- x + alpha*p
        if (LUseO) call daxpy(n_gridpoints,alpha,pp%s,1,usolution%s,1)
        if (LUseT) call daxpy(n_gridpoints,alpha,pp%st,1,usolution%st,1)
        ! b <- b - alpha*q
        if (LUseO) call daxpy(n_gridpoints,-alpha,q%s,1,xb_mg(level,pproc)%s,1)
        if (LUseT) call daxpy(n_gridpoints,-alpha,q%st,1,xb_mg(level,pproc)%st,1)
        ! Calculate norm of residual and exit if it has been
        ! reduced sufficiently
        call start_timer(t_l2norm)
        res_new = l2norm_mnh(xb_mg(level,pproc),.true.)
        call finish_timer(t_l2norm)
        if (mg_param%verbose > 1) then
          if (i_am_master_mpi) then
            write(STDOUT,'(" <MG> ",I7," : ",E10.5,"  ",F10.5)') iter, res_new, res_new/res_old
          end if
        end if
        if (res_new/res_initial < resreduction) then
          solverconverged = .true.
          exit
        end if
        res_old = res_new
        ! Apply preconditioner q <- M^{-1} b
        ! (1) Initialise solution u <- 0
        if (LUseO) xu_mg(level,pproc)%s(:,:,:) = 0.0_rl
        if (LUseT) xu_mg(level,pproc)%st(:,:,:) = 0.0_rl
        ! (2) Call MG Vcycle
        call start_timer(t_prec)
        call mg_vcycle_mnh(xb_mg,xu_mg,xr_mg,finelevel,splitlevel,level,pproc)
        call finish_timer(t_prec)
        ! (3) copy q <- u (solution from MG Vcycle)
        if (LUseO) call dcopy(n_gridpoints,xu_mg(level,pproc)%s,1,q%s,1)
        if (LUseT) call dcopy(n_gridpoints,xu_mg(level,pproc)%st,1,q%st,1)
        call start_timer(t_scalprod)
        call scalarprod_mnh(pproc,q,xb_mg(level,pproc),rz)
        call finish_timer(t_scalprod)
        beta = rz/rz_old
        ! p <- beta*p
        if (LUseO) call dscal(n_gridpoints,beta,pp%s,1)
        if (LUseT) call dscal(n_gridpoints,beta,pp%st,1)
        ! p <- p+q
        if (LUseO) call daxpy(n_gridpoints,1.0_rl,q%s,1,pp%s,1)
        if (LUseT) call daxpy(n_gridpoints,1.0_rl,q%st,1,pp%st,1)
        rz_old = rz
      end do
      call destroy_scalar3d(pp)
      call destroy_scalar3d(q)
    else if (solvertype == SOLVER_RICHARDSON) then
      ! Iterate until convergence
      do iter=1,maxiter
        call start_timer(t_prec)
        call mg_vcycle_mnh(xb_mg,xu_mg,xr_mg,finelevel,splitlevel,level,pproc)
        call finish_timer(t_prec)
        call start_timer(t_res)
        ! Ghosts are up-to-date here, so no need for halo exchange
        call calculate_residual_mnh(level, pproc, &
                                xb_mg(level,pproc),xu_mg(level,pproc),xr_mg(level,pproc))
        call finish_timer(t_res)
        call start_timer(t_l2norm)
        res_new = l2norm_mnh(xr_mg(level,pproc),.true.)
        call finish_timer(t_l2norm)
        if (mg_param%verbose > 1) then
          if (i_am_master_mpi) then
            write(STDOUT,'(" <MG> ",I7," : ",E10.5,"  ",F10.5)') iter, res_new, res_new/res_old
          end if
        end if
        if (res_new/res_initial < resreduction) then
          solverconverged = .true.
          exit
        end if
        res_old = res_new
      end do
      ! Copy u(1) to usolution
      if (LUseO) call dcopy(n_gridpoints,xu_mg(level,pproc)%s,1,usolution%s,1)
      if (LUseT) call dcopy(n_gridpoints,xu_mg(level,pproc)%st,1,usolution%st,1)
    end if
#ifdef MNH_MPPDB
  if ( mppdb_initialized ) then
     call Mppdb_check(      bRHS%st,"mg_solve_mnh:bRHS     "       )
     call Mppdb_check( usolution%st,"mg_solve_mnh:usolution"       )     
  end if
#endif
    
    call destroy_scalar3d(z_one)

    call finish_timer(t_mainloop)

    ! Print out solver information
    if (mg_param%verbose > 0) then
      if (solverconverged) then
        if (i_am_master_mpi) then
          write(STDOUT,'(" <MG> Final residual    ||r|| = ",E12.6)') res_new
          write(STDOUT,'(" <MG> Solver converged in ",I7," iterations rho_{avg} = ",F10.5)') &
          iter, (res_new/res_initial)**(1./(iter))
        end if
      else
        if (i_am_master_mpi) then
          write(STDOUT,'(" <MG> Solver failed to converge after ",I7," iterations rho_{avg} = ",F10.5)') &
          maxiter, (res_new/res_initial)**(1./(iter))
        end if
      end if
    end if
    if (mg_param%verbose >= 100 ) then
    call print_timerinfo("--- Main iteration timing results ---")
    call print_elapsed(t_apply,.True.,1.0_rl)
    call print_elapsed(t_res,.True.,1.0_rl)
    call print_elapsed(t_prec,.True.,1.0_rl)
    call print_elapsed(t_l2norm,.True.,1.0_rl)
    call print_elapsed(t_scalprod,.True.,1.0_rl)
    call print_elapsed(t_mainloop,.True.,1.0_rl)
    end if
!!$    if (i_am_master_mpi) write(STDOUT,'("")')
  end subroutine mg_solve_mnh

  subroutine mg_solve(bRHS,usolution,solver_param)
    implicit none
    type(scalar3d), intent(in)      :: bRHS
    type(scalar3d), intent(inout)   :: usolution
    type(solver_parameters), intent(in) :: solver_param
    integer :: solvertype
    real(kind=rl)  :: resreduction
    integer :: maxiter
    integer :: n_gridpoints
    integer :: iter, level, finelevel, splitlevel
    real(kind=rl) :: res_old, res_new, res_initial
    logical :: solverconverged = .false.
    integer :: nlocalx, nlocaly
    integer :: halo_size
    type(time) :: t_prec, t_res, t_apply, t_l2norm, t_scalprod, t_mainloop
    type(scalar3d) :: pp
    type(scalar3d) :: q
    real(kind=rl) :: alpha, beta, pq, rz, rz_old
    integer :: ierr

    solvertype = solver_param%solvertype
    resreduction = solver_param%resreduction
    maxiter = solver_param%maxiter
    nlocalx = usolution%ix_max - usolution%ix_min+1
    nlocaly = usolution%iy_max - usolution%iy_min+1
    halo_size = usolution%halo_size

    level = mg_param%n_lev
    finelevel = level
    splitlevel = mg_param%lev_split

    ! Initialise timers
    call initialise_timer(t_prec,"t_prec")
    call initialise_timer(t_apply,"t_apply")
    call initialise_timer(t_l2norm,"t_l2norm")
    call initialise_timer(t_scalprod,"t_scalarprod")
    call initialise_timer(t_res,"t_residual")
    call initialise_timer(t_mainloop,"t_mainloop")

    ! Copy b to b(1)
    ! Copy usolution to u(1)
    n_gridpoints = (nlocalx+2*halo_size) &
                 * (nlocaly+2*halo_size) &
                 * (usolution%grid_param%nz+2)
    call dcopy(n_gridpoints,bRHS%s,1,xb_mg(level,pproc)%s,1)
    call dcopy(n_gridpoints,usolution%s,1,xu_mg(level,pproc)%s,1)
! Scale with volume of grid cells
    call volscale_scalar3d(xb_mg(level,pproc),1)
    call start_timer(t_res)
    call calculate_residual(level, pproc, &
                            xb_mg(level,pproc),xu_mg(level,pproc),xr_mg(level,pproc))
    call finish_timer(t_res)
    call start_timer(t_l2norm)
    res_initial = l2norm(xr_mg(level,pproc),.true.)
    call finish_timer(t_l2norm)
    res_old = res_initial
    if (mg_param%verbose > 0) then
      if (i_am_master_mpi) then
        write(STDOUT,'(" *** Multigrid solver ***")')
        write(STDOUT,'(" <MG> Initial residual : ",E10.5)') res_initial
      end if
    end if
    if (mg_param%verbose > 0) then
      if (i_am_master_mpi) then
        write(STDOUT,'(" <MG>   iter :   residual         rho")')
        write(STDOUT,'(" <MG> --------------------------------")')
      end if
    end if

    call mpi_barrier(MPI_COMM_WORLD,ierr)
    call start_timer(t_mainloop)
    if (solvertype == SOLVER_CG) then
      ! NB: b(level,pproc) will be used as r in the following
      call create_scalar3d(MPI_COMM_HORIZ,bRHS%grid_param,halo_size,pp)
      call create_scalar3d(MPI_COMM_HORIZ,bRHS%grid_param,halo_size,q)
      ! Apply preconditioner: Calculate p = M^{-1} r
      ! (1) copy b <- r
      call dcopy(n_gridpoints,xr_mg(level,pproc)%s,1,xb_mg(level,pproc)%s,1)
      ! (2) set u <- 0
      xu_mg(level,pproc)%s(:,:,:) = 0.0_rl
      ! (3) Call MG Vcycle
      call start_timer(t_prec)
      call mg_vcycle(xb_mg,xu_mg,xr_mg,finelevel,splitlevel,level,pproc)
      call finish_timer(t_prec)
      ! (4) copy pp <- u (=solution from MG Vcycle)
      call dcopy(n_gridpoints,xu_mg(level,pproc)%s,1,pp%s,1)
      ! Calculate rz_old = <pp,b>
      call start_timer(t_scalprod)
      call scalarprod(pproc,pp,xb_mg(level,pproc),rz_old)
      call finish_timer(t_scalprod)
      do iter = 1, maxiter
        ! Apply matrix q <- A.pp
        call start_timer(t_apply)
        call apply(pp,q)
        call finish_timer(t_apply)
        ! Calculate pq <- <pp,q>
        call start_timer(t_scalprod)
        call scalarprod(pproc,pp,q,pq)
        call finish_timer(t_scalprod)
        alpha = rz_old/pq
        ! x <- x + alpha*p
        call daxpy(n_gridpoints,alpha,pp%s,1,usolution%s,1)
        ! b <- b - alpha*q
        call daxpy(n_gridpoints,-alpha,q%s,1,xb_mg(level,pproc)%s,1)
        ! Calculate norm of residual and exit if it has been
        ! reduced sufficiently
        call start_timer(t_l2norm)
        res_new = l2norm(xb_mg(level,pproc),.true.)
        call finish_timer(t_l2norm)
        if (mg_param%verbose > 1) then
          if (i_am_master_mpi) then
            write(STDOUT,'(" <MG> ",I7," : ",E10.5,"  ",F10.5)') iter, res_new, res_new/res_old
          end if
        end if
        if (res_new/res_initial < resreduction) then
          solverconverged = .true.
          exit
        end if
        res_old = res_new
        ! Apply preconditioner q <- M^{-1} b
        ! (1) Initialise solution u <- 0
        xu_mg(level,pproc)%s(:,:,:) = 0.0_rl
        ! (2) Call MG Vcycle
        call start_timer(t_prec)
        call mg_vcycle(xb_mg,xu_mg,xr_mg,finelevel,splitlevel,level,pproc)
        call finish_timer(t_prec)
        ! (3) copy q <- u (solution from MG Vcycle)
        call dcopy(n_gridpoints,xu_mg(level,pproc)%s,1,q%s,1)
        call start_timer(t_scalprod)
        call scalarprod(pproc,q,xb_mg(level,pproc),rz)
        call finish_timer(t_scalprod)
        beta = rz/rz_old
        ! p <- beta*p
        call dscal(n_gridpoints,beta,pp%s,1)
        ! p <- p+q
        call daxpy(n_gridpoints,1.0_rl,q%s,1,pp%s,1)
        rz_old = rz
      end do
      call destroy_scalar3d(pp)
      call destroy_scalar3d(q)
    else if (solvertype == SOLVER_RICHARDSON) then
      ! Iterate until convergence
      do iter=1,maxiter
        call start_timer(t_prec)
        call mg_vcycle(xb_mg,xu_mg,xr_mg,finelevel,splitlevel,level,pproc)
        call finish_timer(t_prec)
        call start_timer(t_res)
        ! Ghosts are up-to-date here, so no need for halo exchange
        call calculate_residual(level, pproc, &
                                xb_mg(level,pproc),xu_mg(level,pproc),xr_mg(level,pproc))
        call finish_timer(t_res)
        call start_timer(t_l2norm)
        res_new = l2norm(xr_mg(level,pproc),.true.)
        call finish_timer(t_l2norm)
        if (mg_param%verbose > 1) then
          if (i_am_master_mpi) then
            write(STDOUT,'(" <MG> ",I7," : ",E10.5,"  ",F10.5)') iter, res_new, res_new/res_old
          end if
        end if
        if (res_new/res_initial < resreduction) then
          solverconverged = .true.
          exit
        end if
        res_old = res_new
      end do
      ! Copy u(1) to usolution
      call dcopy(n_gridpoints,xu_mg(level,pproc)%s,1,usolution%s,1)
    end if
    call finish_timer(t_mainloop)

    ! Print out solver information
    if (mg_param%verbose > 0) then
      if (solverconverged) then
        if (i_am_master_mpi) then
          write(STDOUT,'(" <MG> Final residual    ||r|| = ",E12.6)') res_new
          write(STDOUT,'(" <MG> Solver converged in ",I7," iterations rho_{avg} = ",F10.5)') &
          iter, (res_new/res_initial)**(1./(iter))
        end if
      else
        if (i_am_master_mpi) then
          write(STDOUT,'(" <MG> Solver failed to converge after ",I7," iterations rho_{avg} = ",F10.5)') &
          maxiter, (res_new/res_initial)**(1./(iter))
        end if
      end if
    end if
    call print_timerinfo("--- Main iteration timing results ---")
    call print_elapsed(t_apply,.True.,1.0_rl)
    call print_elapsed(t_res,.True.,1.0_rl)
    call print_elapsed(t_prec,.True.,1.0_rl)
    call print_elapsed(t_l2norm,.True.,1.0_rl)
    call print_elapsed(t_scalprod,.True.,1.0_rl)
    call print_elapsed(t_mainloop,.True.,1.0_rl)
    if (i_am_master_mpi) write(STDOUT,'("")')
  end subroutine mg_solve

!==================================================================
! Test haloswap on all levels
!==================================================================
  subroutine measurehaloswap()
    implicit none
    integer :: iter, level, finelevel, splitlevel

    level = mg_param%n_lev
    finelevel = level
    splitlevel = mg_param%lev_split
    call mg_vcyclehaloswaponly(xb_mg,xu_mg,xr_mg,finelevel,splitlevel,level,pproc)
  end subroutine measurehaloswap

end module multigrid

